"""
Lookup dictionary to get values by multiple keys.

Copyright 2024 Vlad Emelianov
"""

import itertools
from collections.abc import Iterator, Mapping
from typing import ClassVar, Generic, TypeVar, cast

from mypy_boto3_builder.constants import ALL

_T = TypeVar("_T")
_V = TypeVar("_V")


class LookupDict(Generic[_V]):
    """
    Lookup dictionary to get values by multiple keys.

    Supports "*" as a wildcard key in a lookup dict.

    Arguments:
        hash_map -- Lookup hash map.
    """

    _ALL: ClassVar = ALL

    def __init__(
        self,
        items: Mapping[str, _T],
    ) -> None:
        self._items = items
        self._lookup_items: dict[tuple[str, ...], _V] = {}
        self._keys_len = 0
        self._products: tuple[tuple[bool, ...], ...] = ()

    @property
    def _lookup(self) -> dict[tuple[str, ...], _V]:
        if not self._lookup_items:
            self._lookup_items = self._generate_lookup({str(k): v for k, v in self._items.items()})
            self._keys_len = len(next(iter(self._lookup_items)))
            self._products = tuple(self._generate_products())

        return self._lookup_items

    def _generate_products(self) -> Iterator[tuple[bool, ...]]:
        static_keys = tuple(
            not any(keys[i] == self._ALL for keys in self._lookup_items)
            for i in range(self._keys_len)
        )
        if all(static_keys):
            yield static_keys
            return

        shifting_keys_count = static_keys.count(False)
        products = itertools.product((True, False), repeat=shifting_keys_count)
        for product in products:
            product_iter = iter(product)
            yield tuple(True if key else next(product_iter) for key in static_keys)

    def _generate_lookup(self, hash_map: Mapping[str, _T]) -> dict[tuple[str, ...], _V]:
        result: dict[tuple[str, ...], _V] = {}
        stack: list[tuple[Mapping[str, _T], tuple[str, ...]]] = [(hash_map, ())]

        while stack:
            current_map, current_keys = stack.pop()
            for key, value in current_map.items():
                new_keys = (*current_keys, key)
                if isinstance(value, Mapping):
                    value_t = cast("Mapping[str, _T]", value)
                    stack.append((value_t, new_keys))
                else:
                    value_v = cast("_V", value)
                    result[new_keys] = value_v

        return result

    def _iterate_lookup_keys(self, keys: tuple[str, ...]) -> Iterator[tuple[str, ...]]:
        if len(keys) != self._keys_len:
            raise ValueError(f"Got {len(keys)}, {self._keys_len} expected: {keys}")

        for product in self._products:
            yield tuple(
                key if is_static else self._ALL
                for is_static, key in zip(product, keys, strict=True)
            )

    def get(self, *keys: str) -> _V | None:
        """
        Get value by multiple keys.
        """
        lookup = self._lookup
        for lookup_keys in self._iterate_lookup_keys(keys):
            result = lookup.get(lookup_keys)
            if result is not None:
                return result
        return None



"""
String to type annotation map that find type annotation by method and argument name.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Mapping
from typing import Final, TypeVar

from mypy_boto3_builder.constants import ALL, CLIENT
from mypy_boto3_builder.import_helpers.import_helper import Import
from mypy_boto3_builder.service_name import ServiceName, ServiceNameCatalog
from mypy_boto3_builder.type_annotations.external_import import ExternalImport
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.type_constant import TypeConstant
from mypy_boto3_builder.type_annotations.type_subscript import TypeSubscript
from mypy_boto3_builder.type_annotations.type_union import TypeUnion
from mypy_boto3_builder.type_maps.literals import QueueAttributeFilterType
from mypy_boto3_builder.type_maps.named_unions import (
    ConditionBaseImportTypeDef,
    CopySourceOrStrTypeDef,
)
from mypy_boto3_builder.type_maps.typed_dicts import CopySourceTypeDef
from mypy_boto3_builder.utils.lookup_dict import LookupDict

__all__ = ("get_method_type_stub",)

_T = TypeVar("_T", bound=FakeAnnotation)
ServiceTypeMap = Mapping[ServiceName, Mapping[str, Mapping[str, Mapping[str, _T]]]]


# Mapping representing default values for botocore methods.
# ServiceName -> TypedDict name -> Argument name -> default value
# Missing value is set to Ellipsis.
DEFAULT_VALUE_MAP: Final[ServiceTypeMap[TypeConstant]] = {
    ServiceNameCatalog.glacier: {
        CLIENT: {
            ALL: {
                "accountId": TypeConstant("-"),
            },
        },
    },
}

_DEFAULT_VALUE_MAP_LOOKUP: LookupDict[TypeConstant] = LookupDict(
    {ServiceNameCatalog.to_str(k): v for k, v in DEFAULT_VALUE_MAP.items()},
)

# Mapping overriding type annotations for botocore method arguments.
# ServiceName -> TypedDict name -> Argument name -> type annotation
# Missing value is parsed from botocore shapes.
TYPE_MAP: Final[ServiceTypeMap[FakeAnnotation]] = {
    ServiceNameCatalog.s3: {
        # FIXME: boto3 calculates ...MD5 for some S3 methods.
        ALL: {
            ALL: {
                "SSECustomerKey": TypeUnion([Type.str, Type.bytes]),
                "CopySourceSSECustomerKey": TypeUnion([Type.str, Type.bytes]),
                "SSECustomerKeyMD5": Type.RemoveArgument,
                "CopySourceSSECustomerKeyMD5": Type.RemoveArgument,
            },
        },
        # FIXME: boto3 overrides CopySource parameters for some S3 methods.
        # Types are set according to docs, might be incorrect
        CLIENT: {
            "copy_object": {"CopySource": CopySourceOrStrTypeDef},
            "upload_part_copy": {"CopySource": CopySourceOrStrTypeDef},
            "copy": {"CopySource": CopySourceTypeDef},
            "put_bucket_acl": {"ContentMD5": Type.RemoveArgument},
            "put_bucket_cors": {"ContentMD5": Type.RemoveArgument},
            "put_bucket_lifecycle": {"ContentMD5": Type.RemoveArgument},
            "put_bucket_logging": {"ContentMD5": Type.RemoveArgument},
            "put_bucket_notification": {"ContentMD5": Type.RemoveArgument},
            "put_bucket_policy": {"ContentMD5": Type.RemoveArgument},
            "put_bucket_replication": {"ContentMD5": Type.RemoveArgument},
            "put_bucket_request_payment": {"ContentMD5": Type.RemoveArgument},
            "put_bucket_tagging": {"ContentMD5": Type.RemoveArgument},
            "put_bucket_versioning": {"ContentMD5": Type.RemoveArgument},
            "put_bucket_website": {"ContentMD5": Type.RemoveArgument},
            "put_object_acl": {"ContentMD5": Type.RemoveArgument},
        },
        "MultipartUploadPart": {
            "copy_from": {"CopySource": CopySourceOrStrTypeDef},
        },
        "Bucket": {"copy": {"CopySource": CopySourceTypeDef}},
        "BucketAcl": {"put": {"ContentMD5": Type.RemoveArgument}},
        "BucketCors": {"put": {"ContentMD5": Type.RemoveArgument}},
        "BucketLifecycle": {"put": {"ContentMD5": Type.RemoveArgument}},
        "BucketLogging": {"put": {"ContentMD5": Type.RemoveArgument}},
        "BucketPolicy": {"put": {"ContentMD5": Type.RemoveArgument}},
        "BucketRequestPayment": {"put": {"ContentMD5": Type.RemoveArgument}},
        "BucketTagging": {"put": {"ContentMD5": Type.RemoveArgument}},
        "BucketVersioning": {
            "enable": {"ContentMD5": Type.RemoveArgument},
            "put": {"ContentMD5": Type.RemoveArgument},
            "suspend": {"ContentMD5": Type.RemoveArgument},
        },
        "BucketWebsite": {"put": {"ContentMD5": Type.RemoveArgument}},
        "Object": {
            "copy": {"CopySource": CopySourceTypeDef},
            "copy_from": {"CopySource": CopySourceOrStrTypeDef},
        },
        "ObjectSummary": {"copy_from": {"CopySource": CopySourceOrStrTypeDef}},
        # FIXME: https://github.com/boto/boto3/issues/3501
        "MultipartUpload": {"Part": {"part_number": Type.int}},
    },
    ServiceNameCatalog.dynamodb: {
        "Table": {
            "batch_writer": {
                "return": ExternalImport(Import.boto3 + "dynamodb" + "table", "BatchWriter")
            },
            "query": {
                "KeyConditionExpression": ConditionBaseImportTypeDef,
                "FilterExpression": ConditionBaseImportTypeDef,
                "ConditionExpression": ConditionBaseImportTypeDef,
            },
            "delete_item": {
                "ConditionExpression": ConditionBaseImportTypeDef,
            },
            "put_item": {
                "ConditionExpression": ConditionBaseImportTypeDef,
            },
            "update_item": {
                "ConditionExpression": ConditionBaseImportTypeDef,
            },
            "scan": {
                "FilterExpression": ConditionBaseImportTypeDef,
            },
        },
    },
    ServiceNameCatalog.sqs: {
        ALL: {
            # FIXME: https://github.com/boto/botocore/issues/2726
            "receive_messages": {
                "AttributeNames": TypeSubscript(Type.Sequence, [QueueAttributeFilterType]),
            },
            "receive_message": {
                "AttributeNames": TypeSubscript(Type.Sequence, [QueueAttributeFilterType]),
            },
            "get_queue_attributes": {
                "AttributeNames": TypeSubscript(Type.Sequence, [QueueAttributeFilterType]),
            },
        },
    },
}

_TYPE_MAP_LOOKUP: LookupDict[FakeAnnotation] = LookupDict(
    {ServiceNameCatalog.to_str(k): v for k, v in TYPE_MAP.items()},
)


def get_method_type_stub(
    service_name: ServiceName,
    class_name: str,
    method_name: str,
    argument_name: str,
) -> FakeAnnotation | None:
    """
    Get stub type for method argument.

    Arguments:
        service_name -- Service name.
        class_name -- Parent class name.
        method_name -- Method name.
        argument_name -- Argument name.

    Returns:
        Type annotation or None.
    """
    return _TYPE_MAP_LOOKUP.get(service_name.name, class_name, method_name, argument_name)


def get_default_value_stub(
    service_name: ServiceName,
    class_name: str,
    method_name: str,
    argument_name: str,
) -> TypeConstant | None:
    """
    Get default value stub for method argument.

    Arguments:
        service_name -- Service name.
        class_name -- Parent class name.
        method_name -- Method name.
        argument_name -- Argument name.

    Returns:
        TypeConstant or None.
    """
    return _DEFAULT_VALUE_MAP_LOOKUP.get(service_name.name, class_name, method_name, argument_name)



"""
Product library choice for CLI.

Copyright 2024 Vlad Emelianov
"""

from enum import Enum


class ProductLibrary(Enum):
    """
    Product library choice for CLI.
    """

    boto3 = "boto3"
    boto3_legacy = "boto3-legacy"
    aiobotocore = "aiobotocore"
    aioboto3 = "aioboto3"
    mypy_boto3 = "mypy_boto3"

    def get_library_name(self) -> str:
        """
        Get library name.
        """
        match self:
            case ProductLibrary.boto3 | ProductLibrary.boto3_legacy | ProductLibrary.mypy_boto3:
                return "boto3"
            case ProductLibrary.aiobotocore:
                return "aiobotocore"
            case ProductLibrary.aioboto3:
                return "aioboto3"

    def get_chat_choice(self) -> str:
        """
        Get chat choice.
        """
        match self:
            case ProductLibrary.boto3 | ProductLibrary.mypy_boto3:
                return "boto3"
            case ProductLibrary.boto3_legacy:
                return "boto3-stubs"
            case ProductLibrary.aiobotocore:
                return "aiobotocore"
            case ProductLibrary.aioboto3:
                return "aioboto3"



"""
Type annotations for aiobotocore.awsrequest module.

Copyright 2025 Vlad Emelianov
"""

from botocore.awsrequest import AWSResponse

class AioAWSResponse(AWSResponse):
    @property
    def content(self) -> bytes: ...
    @property
    def text(self) -> str: ...



"""
Parser that produces `structures.wrapper_package.WrapperPackage`.

Copyright 2024 Vlad Emelianov
"""

from typing import Final

from botocore.config import Config

from mypy_boto3_builder.enums.service_module_name import ServiceModuleName
from mypy_boto3_builder.import_helpers.import_helper import Import
from mypy_boto3_builder.logger import get_logger
from mypy_boto3_builder.parsers.resource_loader import ResourceLoader
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.structures.argument import Argument
from mypy_boto3_builder.structures.function import Function
from mypy_boto3_builder.structures.method import Method
from mypy_boto3_builder.structures.packages.wrapper_package import WrapperPackage
from mypy_boto3_builder.type_annotations.external_import import ExternalImport
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.type_literal import TypeLiteral
from mypy_boto3_builder.type_maps.named_unions import VerifyTypeDef
from mypy_boto3_builder.utils.type_checks import get_optional


class WrapperPackageParser:
    """
    Parser for types-boto3/types-aiobotocore/types-aioboto3 package.

    Arguments:
        package -- Prepared wrapper package with data
    """

    _INIT_ARGUMENTS: Final[tuple[Argument, ...]] = (
        Argument("region_name", get_optional(Type.str), Type.Ellipsis),
        Argument("api_version", get_optional(Type.str), Type.Ellipsis),
        Argument("use_ssl", get_optional(Type.bool), Type.Ellipsis),
        Argument("verify", VerifyTypeDef, Type.Ellipsis),
        Argument("endpoint_url", get_optional(Type.str), Type.Ellipsis),
        Argument("aws_access_key_id", get_optional(Type.str), Type.Ellipsis),
        Argument("aws_secret_access_key", get_optional(Type.str), Type.Ellipsis),
        Argument("aws_session_token", get_optional(Type.str), Type.Ellipsis),
        Argument(
            "config",
            get_optional(ExternalImport.from_class(Config)),
            Type.Ellipsis,
        ),
    )

    def __init__(self, package: WrapperPackage) -> None:
        self.package = package
        self._logger = get_logger()

    def _get_service_name_argument(self, service_name: ServiceName) -> Argument:
        """
        Get service_name argument for init functions.
        """
        return Argument(
            "service_name",
            TypeLiteral(
                f"{service_name.class_name}Type",
                [service_name.boto3_name],
            ),
        )

    def get_client_service_names(self) -> tuple[ServiceName, ...]:
        """
        Get service names that have a Client.
        """
        return self.package.service_names

    def get_resource_service_names(self) -> list[ServiceName]:
        """
        Get service names that have a ServiceResource.
        """
        return ResourceLoader().get_resource_service_names(self.package.service_names)

    def get_init_client_functions(self, name: str = "client") -> list[Function]:
        """
        Extract client data for wrapper package.
        """
        result: list[Function] = []
        client_service_names = self.get_client_service_names()
        add_overload = len(client_service_names) > 1
        decorators = [Type.overload] if add_overload else []
        for service_name in client_service_names:
            package_name = self.package.data.get_service_package_name(service_name)
            service_name_argument = self._get_service_name_argument(service_name)
            return_type = ExternalImport(
                source=Import.from_parts(package_name, ServiceModuleName.client.value),
                name=service_name.get_client_name(),
            )
            arguments = [
                service_name_argument,
                *self._INIT_ARGUMENTS,
            ]
            client_function = Function(
                name=name,
                decorators=decorators,
                arguments=arguments,
                return_type=return_type,
                docstring=f"Create client for {service_name.class_name} service.",
            )
            result.append(client_function)
        return result

    def get_session_client_methods(self, name: str = "client") -> list[Method]:
        """
        Extract client data for wrapper package.
        """
        result: list[Method] = []
        client_service_names = self.get_client_service_names()
        add_overload = len(client_service_names) > 1
        decorators = [Type.overload] if add_overload else []
        for service_name in client_service_names:
            package_name = self.package.data.get_service_package_name(service_name)
            service_name_argument = self._get_service_name_argument(service_name)
            return_type = ExternalImport(
                source=Import.from_parts(package_name, ServiceModuleName.client.value),
                name=service_name.get_client_name(),
            )
            arguments = [
                Argument.self(),
                service_name_argument,
                *self._INIT_ARGUMENTS,
            ]
            method = Method(
                name=name,
                decorators=decorators,
                arguments=arguments,
                return_type=return_type,
                docstring=f"Create client for {service_name.class_name} service.",
            )
            result.append(method)
        return result

    def get_init_resource_functions(self, name: str = "resource") -> list[Function]:
        """
        Extract resource data for wrapper package.
        """
        result: list[Function] = []
        resource_service_names = self.get_resource_service_names()
        add_overload = len(resource_service_names) > 1
        decorators = [Type.overload] if add_overload else []
        for service_name in resource_service_names:
            package_name = self.package.data.get_service_package_name(service_name)
            service_name_argument = self._get_service_name_argument(service_name)
            return_type = ExternalImport(
                source=Import.from_parts(package_name, ServiceModuleName.service_resource.value),
                name=service_name.get_service_resource_name(),
            )
            arguments = [
                service_name_argument,
                *self._INIT_ARGUMENTS,
            ]
            resource_function = Function(
                name=name,
                decorators=decorators,
                arguments=arguments,
                return_type=return_type,
                docstring=f"Create ServiceResource for {service_name.class_name} service.",
            )
            result.append(resource_function)
        return result

    def get_session_resource_methods(self, name: str = "resource") -> list[Method]:
        """
        Extract resource data for wrapper package.
        """
        result: list[Method] = []
        resource_service_names = self.get_resource_service_names()
        add_overload = len(resource_service_names) > 1
        decorators = [Type.overload] if add_overload else []
        for service_name in resource_service_names:
            package_name = self.package.data.get_service_package_name(service_name)
            service_name_argument = self._get_service_name_argument(service_name)
            return_type = ExternalImport(
                source=Import.from_parts(package_name, ServiceModuleName.service_resource.value),
                name=service_name.get_service_resource_name(),
            )
            arguments = [
                Argument.self(),
                service_name_argument,
                *self._INIT_ARGUMENTS,
            ]
            resource_method = Method(
                name=name,
                decorators=decorators,
                arguments=arguments,
                return_type=return_type,
                docstring=f"Create ServiceResource for {service_name.class_name} service.",
            )
            result.append(resource_method)
        return result



"""
Getters for boto3 client and resource from session.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Iterable
from functools import cache

from botocore.session import Session as BotocoreSession
from botocore.session import get_session

from mypy_boto3_builder.service_name import ServiceName, ServiceNameCatalog
from mypy_boto3_builder.type_annotations.type_literal import TypeLiteral
from mypy_boto3_builder.utils.strings import get_botocore_class_name


@cache
def get_botocore_session() -> BotocoreSession:
    """
    Get botocore session from boto3 session.
    """
    session = get_session()
    session.set_credentials("access_key", "secret_key", "token")
    return session


def get_region_name_literal(
    service_names: Iterable[ServiceName],
) -> TypeLiteral | None:
    """
    Get Literal with all regions.

    Arguments:
        service_names -- All available service names.

    Returns:
        TypeLiteral for region names.
    """
    session = get_botocore_session()
    children: set[str] = set()
    for service_name in service_names:
        children.update(session.get_available_regions(service_name.boto3_name))
    if not children:
        return None
    return TypeLiteral("RegionName", children)


def get_available_service_names() -> list[ServiceName]:
    """
    Get a list of botocore supported service names.

    Returns:
        A list of supported services.
    """
    session = get_botocore_session()
    available_services = session.get_available_services()
    result: list[ServiceName] = []
    for name in available_services:
        service_data = session.get_service_data(name)
        metadata = service_data["metadata"]
        class_name = get_botocore_class_name(metadata)
        service_name = ServiceNameCatalog.add(name, class_name)
        result.append(service_name)
    return result



"""
Chat enums.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Iterable
from enum import Enum
from pathlib import Path

from mypy_boto3_builder.chat.text_style import TextStyle
from mypy_boto3_builder.chat.type_defs import Message
from mypy_boto3_builder.enums.product import Product
from mypy_boto3_builder.enums.product_library import ProductLibrary
from mypy_boto3_builder.package_data import (
    Boto3StubsCustomPackageData,
    TypesAioBoto3CustomPackageData,
    TypesAioBotocoreCustomPackageData,
    TypesBoto3CustomPackageData,
)


class ServiceActions(Enum):
    """
    Available service actions.
    """

    build = "Continue"
    add = "Add another service"
    add_all = "Add all available services"
    recommended = "Add only recommended services"
    remove = "Remove selected service"
    remove_all = "Remove all selected services"


class PackageManager(Enum):
    """
    Available package managers.
    """

    pip = "pip"
    uv = "uv"
    poetry = "poetry"
    pipenv = "pipenv"

    @property
    def url(self) -> str:
        """
        Get package manager URL.
        """
        match self:
            case PackageManager.pip:
                return "https://pypi.org/project/pip/"
            case PackageManager.uv:
                return "https://docs.astral.sh/uv/"
            case PackageManager.poetry:
                return "https://python-poetry.org/"
            case PackageManager.pipenv:
                return "https://pipenv.pypa.io/en/latest/"

    def get_install_cmd(self, whl_path: Path) -> str:
        """
        Get install command.
        """
        match self:
            case PackageManager.uv:
                return f"uv add --dev {whl_path}"
            case PackageManager.pip:
                return f"pip install {whl_path}"
            case PackageManager.poetry:
                return f"poetry add -G dev {whl_path}"
            case PackageManager.pipenv:
                return f"pipenv install --dev {whl_path}"


class Library(Enum):
    """
    Library type.
    """

    boto3 = "boto3"
    aiobotocore = "aiobotocore"
    aioboto3 = "aioboto3"
    boto3_legacy = "boto3-stubs"

    @property
    def product_library(self) -> ProductLibrary:
        """
        Get product library.
        """
        match self:
            case Library.boto3:
                return ProductLibrary.boto3
            case Library.aiobotocore:
                return ProductLibrary.aiobotocore
            case Library.aioboto3:
                return ProductLibrary.aioboto3
            case Library.boto3_legacy:
                return ProductLibrary.boto3_legacy

    @property
    def text(self) -> Message:
        """
        Get choice user text.
        """
        match self:
            case Library.boto3_legacy:
                return [
                    "boto3",
                    TextStyle.text.wrap(", but it also already has "),
                    TextStyle.tag.wrap("boto3-stubs"),
                ]
            case _:
                return [self.value]

    @property
    def title(self) -> Message:
        """
        Get choice title.
        """
        return [f"{self.value:<14}"]

    @property
    def title_info(self) -> str:
        """
        Get choice title.
        """
        match self:
            case Library.boto3:
                return "AWS SDK for Python"
            case Library.aiobotocore:
                return "Async client for amazon services using botocore and aiohttp/asyncio."
            case Library.aioboto3:
                return "Async AWS SDK for Python"
            case Library.boto3_legacy:
                return "Legacy type annotations for AWS SDK for Python"

    @property
    def documentation_url(self) -> str:
        """
        Link to documentation.
        """
        match self:
            case Library.boto3:
                return TypesBoto3CustomPackageData.local_doc_link
            case Library.boto3_legacy:
                return Boto3StubsCustomPackageData.local_doc_link
            case Library.aiobotocore:
                return TypesAioBotocoreCustomPackageData.local_doc_link
            case Library.aioboto3:
                return TypesAioBoto3CustomPackageData.local_doc_link

    def get_package_prefix(self) -> str:
        """
        Get *.whl package name prefix.
        """
        match self:
            case Library.boto3:
                return "types_boto3"
            case Library.boto3_legacy:
                return "boto3_stubs"
            case Library.aiobotocore:
                return "types_aiobotocore"
            case Library.aioboto3:
                return "types_aioboto3"

    @property
    def product(self) -> Product:
        """
        Get product.
        """
        match self:
            case Library.boto3:
                return Product.types_boto3_custom
            case Library.boto3_legacy:
                return Product.boto3_stubs_custom
            case Library.aiobotocore:
                return Product.types_aiobotocore_custom
            case Library.aioboto3:
                return Product.types_aioboto3_custom

    @classmethod
    def from_products(cls, products: Iterable[Product]) -> "Library | None":
        """
        Get library from product.
        """
        for product in products:
            match product:
                case Product.types_boto3_custom:
                    return Library.boto3
                case Product.boto3_stubs_custom:
                    return Library.boto3_legacy
                case Product.types_aiobotocore_custom:
                    return Library.aiobotocore
                case Product.types_aioboto3_custom:
                    return Library.aioboto3
                case _:
                    pass
        return None



"""
Structure for types-aioboto3 module.

Copyright 2024 Vlad Emelianov
"""

from mypy_boto3_builder.structures.packages.wrapper_package import WrapperPackage
from mypy_boto3_builder.utils.version import get_max_build_version, get_min_build_version
from mypy_boto3_builder.utils.version_getters import get_aiobotocore_version


class TypesAioBoto3Package(WrapperPackage):
    """
    Structure for types-aioboto3 module.
    """

    def get_all_names(self) -> list[str]:
        """
        Get names for `__all__` directive.
        """
        result = [
            "Session",
        ]
        return sorted(result)

    @property
    def min_library_version(self) -> str:
        """
        Minimum required library version.
        """
        return get_min_build_version(get_aiobotocore_version())

    @property
    def max_library_version(self) -> str:
        """
        Minimum required library version.
        """
        return get_max_build_version(get_aiobotocore_version())



"""
Parser for Boto3 ServiceResource identifiers, produces `structures.Attribute`.

Copyright 2024 Vlad Emelianov
"""

from typing import TYPE_CHECKING

from mypy_boto3_builder.boto3_ports.model import ResourceModel
from mypy_boto3_builder.structures.attribute import Attribute
from mypy_boto3_builder.type_annotations.internal_import import InternalImport
from mypy_boto3_builder.type_annotations.type import Type

if TYPE_CHECKING:
    from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation


def parse_references(resource_model: ResourceModel) -> list[Attribute]:
    """
    Extract references from boto3 resource.

    Arguments:
        resource_model -- boto3 ResourceModel.

    Returns:
        A list of Attribute structures.
    """
    result: list[Attribute] = []
    for reference in resource_model.references:
        if not reference.resource:
            continue
        type_annotation: FakeAnnotation = InternalImport(reference.resource.type)
        if reference.resource.path and "[]" in reference.resource.path:
            type_annotation = Type.wrap_list(type_annotation)
        result.append(Attribute(reference.name, type_annotation=type_annotation, is_reference=True))
    return result



"""
Type annotations for aioboto3.s3.cse module.

Copyright 2025 Vlad Emelianov
"""

import asyncio
from types import TracebackType
from typing import IO, Any, Mapping, TypeVar

from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey

_R = TypeVar("_R")

RANGE_REGEX: Any
AES_BLOCK_SIZE: int
AES_BLOCK_SIZE_BYTES: int
JAVA_LONG_MAX_VALUE: int

class DummyAIOFile:
    file: IO[Any]
    def __init__(self, data: bytes) -> None: ...
    async def read(self, n: int = ...) -> bytes: ...
    async def readany(self) -> bytes: ...
    async def readexactly(self, n: int) -> bytes: ...
    async def readchunk(self) -> bytes: ...

class DecryptError(Exception): ...

class CryptoContext:
    async def setup(self) -> None: ...
    async def close(self) -> None: ...
    async def get_decryption_aes_key(
        self, key: bytes, material_description: dict[str, Any]
    ) -> bytes: ...
    async def get_encryption_aes_key(self) -> tuple[bytes, dict[str, str], str]: ...

class AsymmetricCryptoContext(CryptoContext):
    public_key: RSAPublicKey
    private_key: RSAPrivateKey
    def __init__(
        self,
        public_key: RSAPublicKey | None = ...,
        private_key: RSAPrivateKey | None = ...,
        loop: asyncio.AbstractEventLoop | None = ...,
    ) -> None: ...
    async def get_decryption_aes_key(
        self, key: bytes, material_description: dict[str, Any]
    ) -> bytes: ...
    async def get_encryption_aes_key(self) -> tuple[bytes, dict[str, str], str]: ...
    @staticmethod
    def from_der_public_key(data: bytes) -> RSAPublicKey: ...
    @staticmethod
    def from_der_private_key(data: bytes, password: str | None = ...) -> RSAPrivateKey: ...

class SymmetricCryptoContext(CryptoContext):
    key: bytes
    def __init__(self, key: bytes, loop: asyncio.AbstractEventLoop | None = ...) -> None: ...
    async def get_decryption_aes_key(
        self, key: bytes, material_description: dict[str, Any]
    ) -> bytes: ...
    async def get_encryption_aes_key(self) -> tuple[bytes, dict[str, str], str]: ...

class KMSCryptoContext(CryptoContext):
    kms_key: str
    authenticated_encryption: bool
    def __init__(
        self,
        keyid: str | None = ...,
        kms_client_args: dict[str, Any] | None = ...,
        authenticated_encryption: bool = ...,
    ) -> None: ...
    async def setup(self) -> None: ...
    async def close(self) -> None: ...
    async def get_decryption_aes_key(
        self, key: bytes, material_description: dict[str, Any]
    ) -> bytes: ...
    async def get_encryption_aes_key(self) -> tuple[bytes, dict[str, str], str]: ...

class MockKMSCryptoContext(KMSCryptoContext):
    aes_key: bytes
    material_description: dict[str, Any]
    encrypted_key: bytes
    authenticated_encryption: bool
    def __init__(
        self,
        aes_key: bytes,
        material_description: dict[str, Any],
        encrypted_key: bytes,
        authenticated_encryption: bool = ...,
    ) -> None: ...
    async def setup(self) -> None: ...
    async def close(self) -> None: ...
    async def get_decryption_aes_key(
        self, key: bytes, material_description: dict[str, Any]
    ) -> bytes: ...
    async def get_encryption_aes_key(self) -> tuple[bytes, dict[str, str], str]: ...

class S3CSE:
    def __init__(
        self, crypto_context: CryptoContext, s3_client_args: dict[str, Any] | None = ...
    ) -> None: ...
    async def setup(self) -> None: ...
    async def close(self) -> None: ...
    async def __aenter__(self: _R) -> _R: ...
    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        tb: TracebackType | None,
    ) -> None: ...
    async def get_object(self, Bucket: str, Key: str, **kwargs: Any) -> dict[str, Any]: ...
    async def put_object(
        self,
        Body: bytes | IO[Any],
        Bucket: str,
        Key: str,
        Metadata: Mapping[str, Any] | None = ...,
        **kwargs: Any,
    ) -> Any: ...



"""
Type annotations for aiobotocore.response module.

Copyright 2025 Vlad Emelianov
"""

import asyncio
from types import TracebackType
from typing import IO, Any, AsyncIterator, TypeVar

import wrapt  # type: ignore
from aiobotocore import parsers as parsers
from botocore.exceptions import ReadTimeoutError
from botocore.model import OperationModel
from requests.models import Response

_R = TypeVar("_R")

class AioReadTimeoutError(ReadTimeoutError, asyncio.TimeoutError): ...

class StreamingBody(wrapt.ObjectProxy):  # type: ignore
    def __init__(self, raw_stream: IO[bytes], content_length: int) -> None: ...
    async def __aenter__(self: _R) -> _R: ...
    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        tb: TracebackType | None,
    ) -> None: ...
    def tell(self) -> int: ...
    def readable(self) -> bool: ...
    async def read(self, amt: int | None = ...) -> bytes: ...
    async def readlines(self) -> list[bytes]: ...
    def __aiter__(self) -> AsyncIterator[bytes]: ...
    async def __anext__(self) -> bytes: ...
    anext: Any
    def iter_lines(self, chunk_size: int = ..., keepends: bool = ...) -> AsyncIterator[bytes]: ...
    def iter_chunks(self, chunk_size: int = ...) -> AsyncIterator[bytes]: ...

async def get_response(
    operation_model: OperationModel, http_response: Response
) -> tuple[Response, Any]: ...



"""
Product choice for CLI.

Copyright 2024 Vlad Emelianov
"""

from enum import Enum

from mypy_boto3_builder.enums.product_library import ProductLibrary
from mypy_boto3_builder.enums.product_type import ProductType


class Product(Enum):
    """
    Product choice for CLI.
    """

    types_boto3 = "types-boto3"
    types_boto3_lite = "types-boto3-lite"
    types_boto3_services = "types-boto3-services"
    types_boto3_full = "types-boto3-full"
    types_boto3_custom = "types-boto3-custom"
    types_boto3_docs = "types-boto3-docs"

    types_aiobotocore = "aiobotocore"
    types_aiobotocore_lite = "aiobotocore-lite"
    types_aiobotocore_services = "aiobotocore-services"
    types_aiobotocore_full = "aiobotocore-full"
    types_aiobotocore_custom = "aiobotocore-custom"
    types_aiobotocore_docs = "aiobotocore-docs"

    types_aioboto3 = "aioboto3"
    types_aioboto3_lite = "aioboto3-lite"
    types_aioboto3_custom = "aioboto3-custom"
    types_aioboto3_docs = "aioboto3-docs"

    boto3_stubs = "boto3"
    boto3_stubs_lite = "boto3-lite"
    boto3_stubs_services = "boto3-services"
    boto3_stubs_full = "boto3-full"
    boto3_stubs_custom = "boto3-custom"
    boto3_stubs_docs = "boto3-docs"

    mypy_boto3 = "mypy-boto3"

    def __str__(self) -> str:
        """
        Get string representation for debugging.
        """
        return self.value

    def get_library(self) -> ProductLibrary:
        """
        Get library name.
        """
        match self:
            case (
                Product.types_boto3
                | Product.types_boto3_lite
                | Product.types_boto3_services
                | Product.types_boto3_full
                | Product.types_boto3_docs
                | Product.types_boto3_custom
            ):
                return ProductLibrary.boto3
            case (
                Product.boto3_stubs
                | Product.boto3_stubs_lite
                | Product.boto3_stubs_services
                | Product.boto3_stubs_full
                | Product.boto3_stubs_docs
                | Product.boto3_stubs_custom
            ):
                return ProductLibrary.boto3_legacy
            case (
                Product.types_aiobotocore
                | Product.types_aiobotocore_lite
                | Product.types_aiobotocore_services
                | Product.types_aiobotocore_full
                | Product.types_aiobotocore_docs
                | Product.types_aiobotocore_custom
            ):
                return ProductLibrary.aiobotocore
            case (
                Product.types_aioboto3
                | Product.types_aioboto3_lite
                | Product.types_aioboto3_docs
                | Product.types_aioboto3_custom
            ):
                return ProductLibrary.aioboto3
            case Product.mypy_boto3:
                return ProductLibrary.mypy_boto3

    def get_type(self) -> ProductType:
        """
        Get product type.
        """
        match self:
            case (
                Product.types_boto3
                | Product.boto3_stubs
                | Product.types_aiobotocore
                | Product.types_aioboto3
                | Product.mypy_boto3
            ):
                return ProductType.stubs
            case (
                Product.types_boto3_lite
                | Product.boto3_stubs_lite
                | Product.types_aiobotocore_lite
                | Product.types_aioboto3_lite
            ):
                return ProductType.stubs_lite
            case (
                Product.types_boto3_services
                | Product.boto3_stubs_services
                | Product.types_aiobotocore_services
            ):
                return ProductType.service_stubs
            case (
                Product.types_boto3_docs
                | Product.boto3_stubs_docs
                | Product.types_aiobotocore_docs
                | Product.types_aioboto3_docs
            ):
                return ProductType.docs
            case (
                Product.types_boto3_full | Product.boto3_stubs_full | Product.types_aiobotocore_full
            ):
                return ProductType.full
            case (
                Product.types_boto3_custom
                | Product.boto3_stubs_custom
                | Product.types_aiobotocore_custom
                | Product.types_aioboto3_custom
            ):
                return ProductType.custom



"""
Structure for types-aiobotocore module.

Copyright 2024 Vlad Emelianov
"""

from mypy_boto3_builder.structures.packages.wrapper_package import WrapperPackage


class TypesAioBotocorePackage(WrapperPackage):
    """
    Structure for types-aiobotocore module.
    """

    def get_all_names(self) -> list[str]:
        """
        Get names for `__all__` directive.
        """
        result = [
            "get_session",
            "Session",
        ]
        return sorted(result)



"""
Parent class for all package structures.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Iterable
from typing import TYPE_CHECKING

from mypy_boto3_builder.constants import SUPPORTED_PY_VERSIONS
from mypy_boto3_builder.exceptions import StructureError
from mypy_boto3_builder.logger import get_logger
from mypy_boto3_builder.package_data import BasePackageData
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.structures.package_url import PackageURL
from mypy_boto3_builder.utils.install_requires import InstallRequires
from mypy_boto3_builder.utils.version import (
    get_max_build_version,
    get_min_build_version,
    is_valid_version,
)
from mypy_boto3_builder.utils.version_getters import get_botocore_version

if TYPE_CHECKING:
    from mypy_boto3_builder.structures.package_extra import PackageExtra


class Package:
    """
    Parent class for all package structures.
    """

    def __init__(
        self,
        data: BasePackageData,
        service_names: Iterable[ServiceName] = (),
        version: str = "",
    ) -> None:
        self.data = data
        self._pypi_name = self.data.pypi_name
        self._version: str = ""
        if version:
            self.version = version
        self.service_names = tuple(service_names)
        self.logger = get_logger()
        self.url = PackageURL(self.pypi_name, self.data)
        self.extras: list[PackageExtra] = []
        self.install_requires = InstallRequires(self.data.install_requires)

    @property
    def library_version(self) -> str:
        """
        Library version.
        """
        return self.data.get_library_version()

    @property
    def botocore_version(self) -> str:
        """
        Botocore version.
        """
        return get_botocore_version()

    @property
    def pypi_name(self) -> str:
        """
        PyPI package name.
        """
        return self._pypi_name

    @pypi_name.setter
    def pypi_name(self, value: str) -> None:
        self._pypi_name = value
        self.url.pypi_name = value

    @property
    def version(self) -> str:
        """
        Package version.
        """
        if not self._version:
            raise StructureError(f"Version is not set for {self.pypi_name}")
        return self._version

    @version.setter
    def version(self, value: str) -> None:
        if not is_valid_version(value):
            raise StructureError(f"Invalid version: {value}") from None
        self._version = value

    @property
    def name(self) -> str:
        """
        Package name.
        """
        if not self.data.name:
            raise StructureError(f"Package name is not set for {self.pypi_name}")

        return self.data.name

    @property
    def library_name(self) -> str:
        """
        PyPI library package name.
        """
        return self.data.get_library_name()

    @property
    def library_chat_choice(self) -> str:
        """
        Package library chat choice.
        """
        return self.data.get_library_chat_choice()

    def has_main_package(self) -> bool:
        """
        Check if package has main package.
        """
        return bool(self.data.name)

    @property
    def service_name(self) -> ServiceName:
        """
        Service name for the package.
        """
        if len(self.service_names) != 1:
            raise StructureError(f"Package {self.name} has more than one service name")
        return self.service_names[0]

    @property
    def directory_name(self) -> str:
        """
        Directory name to store generated package.
        """
        underscore_package_name = self.pypi_name.replace("-", "_")
        return f"{underscore_package_name}_package"

    def __str__(self) -> str:
        """
        Get string representation for debugging.
        """
        return f"{self.name} {self._version} ({self.library_name} {self.library_version})"

    def get_local_doc_link(self) -> str:
        """
        Get link to local docs.
        """
        return self.data.local_doc_link

    def get_service_local_doc_link(self, service_name: ServiceName) -> str:
        """
        Get link to service local docs.
        """
        url = self.data.local_doc_link
        if service_name:
            url = f"{url}{self.get_module_name(service_name)}/"
        return url

    def get_module_name(self, service_name: ServiceName) -> str:
        """
        Get service module name.
        """
        return self.data.get_service_package_name(service_name)

    def get_service_pypi_name(self, service_name: ServiceName) -> str:
        """
        Get PyPI package name for a service package.
        """
        return self.data.get_service_pypi_name(service_name)

    @property
    def min_library_version(self) -> str:
        """
        Minimum required library version.
        """
        return get_min_build_version(self.library_version)

    @property
    def max_library_version(self) -> str:
        """
        Minimum required library version.
        """
        return get_max_build_version(self.library_version)

    @property
    def min_python_version(self) -> str:
        """
        Minimum required python version.
        """
        min_version = min(SUPPORTED_PY_VERSIONS)
        return ".".join(str(i) for i in min_version)

    def get_classifiers(self) -> list[str]:
        """
        Get classifiers for package.
        """
        result = [
            "Development Status :: 5 - Production/Stable",
            "Intended Audience :: Developers",
            "Environment :: Console",
            "License :: OSI Approved :: MIT License",
            "Natural Language :: English",
            "Operating System :: OS Independent",
        ]
        major_versions = {version[0] for version in SUPPORTED_PY_VERSIONS}
        for major in sorted(major_versions):
            result.append(f"Programming Language :: Python :: {major}")
            minor_versions = {
                version[1]
                for version in SUPPORTED_PY_VERSIONS
                if version[0] == major and len(version) > 1
            }
            result.extend(
                f"Programming Language :: Python :: {major}.{minor}"
                for minor in sorted(minor_versions)
            )
        if len(major_versions) == 1:
            major = next(iter(major_versions))
            result.append(f"Programming Language :: Python :: {major} :: Only")
        result.extend(
            (
                "Programming Language :: Python :: Implementation :: CPython",
                "Typing :: Stubs Only",
            ),
        )
        return result

    @property
    def essential_service_names(self) -> list[ServiceName]:
        """
        List of services maked as essential.
        """
        return [service_name for service_name in self.service_names if service_name.is_essential()]



"""
Boto3 ServiceResource.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Generator

from mypy_boto3_builder.enums.service_module_name import ServiceModuleName
from mypy_boto3_builder.exceptions import StructureError
from mypy_boto3_builder.import_helpers.import_helper import Import
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.structures.attribute import Attribute
from mypy_boto3_builder.structures.class_record import ClassRecord
from mypy_boto3_builder.structures.collection import Collection
from mypy_boto3_builder.structures.resource_record import ResourceRecord
from mypy_boto3_builder.type_annotations.external_import import ExternalImport
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation


class ServiceResource(ClassRecord):
    """
    Service Resource.
    """

    def __init__(
        self,
        name: str,
        service_name: ServiceName,
    ) -> None:
        super().__init__(
            name=name,
            bases=[ExternalImport(Import.boto3 + "resources" + "base", "ServiceResource")],
        )
        self.service_name = service_name
        self.collections: list[Collection] = []
        self.sub_resources: list[ResourceRecord] = []
        self.resource_meta_class = self._get_resource_meta_class()

    def __hash__(self) -> int:
        """
        Calculate hash value based on service name.
        """
        return hash(self.service_name)

    @property
    def alias_name(self) -> str:
        """
        Class alias name for safe import.
        """
        return "ServiceResource"

    @property
    def boto3_doc_link(self) -> str:
        """
        Link to boto3 docs.
        """
        return f"{self.service_name.boto3_doc_link_parent}/service-resource/index.html"

    def iterate_types(self) -> Generator[FakeAnnotation]:
        """
        Iterate over type annotations for collections and sub-resources.
        """
        yield from super().iterate_types()
        yield from self.resource_meta_class.iterate_types()
        for collection in self.collections:
            yield from collection.iterate_types()
        for sub_resource in self.sub_resources:
            yield from sub_resource.iterate_types()

    def get_all_names(self) -> list[str]:
        """
        Get names for `__all__` statement.
        """
        result = [self.name]
        result.extend(resource.name for resource in self.sub_resources)
        result.extend(collection.name for collection in self.get_collections())
        return sorted(result)

    def get_collections(self) -> list[Collection]:
        """
        Get a list of Service Resource collections.
        """
        collection_names = [i.name for i in self.collections]
        result: list[Collection] = []
        result.extend(self.collections)
        for resource in self.sub_resources:
            for collection in resource.collections:
                if collection.name in collection_names:
                    raise StructureError(f"Conflicting collections: {collection.name}")
                collection_names.append(collection.name)
                result.append(collection)

        return result

    def get_sub_resources(self) -> tuple[ResourceRecord, ...]:
        """
        Get sub-resources.

        Returns:
            A list of sub resources.
        """
        return tuple(self.sub_resources)

    def _get_resource_meta_class(self) -> ClassRecord:
        return ClassRecord(
            name=f"{self.service_name.class_name}ResourceMeta",
            bases=[ExternalImport(Import.boto3 + "resources" + "base", "ResourceMeta")],
            attributes=[
                Attribute(
                    "client",
                    ExternalImport(
                        source=Import.local(ServiceModuleName.client.value),
                        name=self.service_name.get_client_name(),
                    ),
                    type_ignore="override",
                ),
            ],
        )



"""
Type annotations for Chat.

Copyright 2024 Vlad Emelianov
"""

from __future__ import annotations

from collections.abc import MutableSequence

from mypy_boto3_builder.chat.text_style import TextStyle

MessagePair = tuple[TextStyle, str]
MessageToken = MessagePair | str
Message = tuple[MessageToken, ...] | MutableSequence[MessageToken]



"""
InstallRequires container.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Iterable, Iterator
from dataclasses import dataclass

from mypy_boto3_builder.utils.version import sort_versions


@dataclass
class InstallRequiresItem:
    """
    Item for install_requires.
    """

    name: str
    drop_python_version: str | None = None
    min_version: str | None = None

    def __hash__(self) -> int:
        """
        Calculate hash value.
        """
        return hash(self.name)

    @classmethod
    def parse(cls, item: str) -> "InstallRequiresItem":
        """
        Parse item from string.
        """
        name = item
        min_version: str | None = None
        drop_python_version: str | None = None
        if ">=" in name:
            name, min_version = name.split(">=")

        if "; python_version<" in name:
            name, drop_python_version = name.split("; python_version<")
            drop_python_version = drop_python_version.replace('"', "")

        return cls(name=name, min_version=min_version, drop_python_version=drop_python_version)

    def render(self) -> str:
        """
        Render item to a valid install_requires string for setup.py.
        """
        version = f">={self.min_version}" if self.min_version else ""

        python_version = ""
        if self.drop_python_version:
            python_version = f'; python_version<"{self.drop_python_version}"'

        return f"{self.name}{version}{python_version}"


class InstallRequires:
    """
    InstallRequires container.
    """

    def __init__(self, items: Iterable[InstallRequiresItem | str] = ()) -> None:
        self.items: list[InstallRequiresItem] = []
        self.add(*items)

    def __bool__(self) -> bool:
        """
        Whether install_requires is not empty.
        """
        return bool(self.items)

    def get(self, name: str) -> InstallRequiresItem | None:
        """
        Get item by name.
        """
        for item in self.items:
            if item.name == name:
                return item
        return None

    def update(self, other: "InstallRequires") -> None:
        """
        Update install_requires with another one.
        """
        for item in other.items:
            self.add(item)

    def add(self, *raw_items: InstallRequiresItem | str) -> None:
        """
        Add item to install_requires.
        """
        for raw_item in raw_items:
            item = InstallRequiresItem.parse(raw_item) if isinstance(raw_item, str) else raw_item
            old_item = self.get(item.name)
            if not old_item:
                self.items.append(item)
                continue

            drop_python_versions = sort_versions(
                filter(None, (item.drop_python_version, old_item.drop_python_version))
            )
            min_versions = sort_versions(filter(None, (item.min_version, old_item.min_version)))

            old_item.drop_python_version = (
                drop_python_versions[-1] if drop_python_versions else None
            )
            old_item.min_version = min_versions[-1] if min_versions else None

    def iterate_items(self) -> Iterator[InstallRequiresItem]:
        """
        Iterate over all items for rendering.
        """
        yield from sorted(self.items, key=lambda x: x.name)

    def clear(self) -> None:
        """
        Clear install_requires.
        """
        self.items = []



"""
Type annotations for aioboto3.resources.collection module.

Copyright 2025 Vlad Emelianov
"""

import logging
from typing import Any, AsyncIterator, NoReturn

from boto3.resources.base import ServiceResource
from boto3.resources.collection import CollectionFactory, CollectionManager, ResourceCollection
from boto3.resources.factory import ResourceFactory
from boto3.resources.model import Collection
from boto3.utils import ServiceContext
from botocore.hooks import BaseEventHooks

logger: logging.Logger

class AIOResourceCollection(ResourceCollection):
    def __anext__(self) -> AsyncIterator[Any]: ...
    def __aiter__(self) -> AsyncIterator[Any]: ...
    def __iter__(self) -> NoReturn: ...
    def pages(self) -> AsyncIterator[list[Any]]: ...  # type: ignore[override]

class AIOCollectionManager(CollectionManager):
    def __init__(
        self,
        collection_model: Collection,
        parent: ServiceResource,
        factory: ResourceFactory,
        service_context: ServiceContext,
    ) -> None: ...

class AIOCollectionFactory(CollectionFactory):
    def load_from_definition(
        self,
        resource_name: str,
        collection_model: Collection,
        service_context: ServiceContext,
        event_emitter: BaseEventHooks,
    ) -> AIOCollectionManager: ...



"""
Type annotations for aiobotocore.hooks module.

Copyright 2025 Vlad Emelianov
"""

from typing import Any

from botocore.hooks import HierarchicalEmitter
from requests.models import Response

class AioHierarchicalEmitter(HierarchicalEmitter):
    async def emit_until_response(self, event_name: str, **kwargs: Any) -> list[Response]: ...



"""
Collection of Literals added by boto3.

Copyright 2024 Vlad Emelianov
"""

from mypy_boto3_builder.type_annotations.type_literal import TypeLiteral

QueueAttributeFilterType = TypeLiteral(
    "QueueAttributeFilterType",
    (
        "All",
        "AWSTraceHeader",
        "ApproximateFirstReceiveTimestamp",
        "ApproximateReceiveCount",
        "MessageDeduplicationId",
        "MessageGroupId",
        "SenderId",
        "SentTimestamp",
        "SequenceNumber",
        "Policy",
        "VisibilityTimeout",
        "MaximumMessageSize",
        "MessageRetentionPeriod",
        "ApproximateNumberOfMessages",
        "ApproximateNumberOfMessagesNotVisible",
        "CreatedTimestamp",
        "LastModifiedTimestamp",
        "QueueArn",
        "ApproximateNumberOfMessagesDelayed",
        "DelaySeconds",
        "ReceiveMessageWaitTimeSeconds",
        "RedrivePolicy",
        "FifoQueue",
        "ContentBasedDeduplication",
        "KmsMasterKeyId",
        "KmsDataKeyReusePeriodSeconds",
        "DeduplicationScope",
        "FifoThroughputLimit",
        "RedriveAllowPolicy",
        "SqsManagedSseEnabled",
    ),
)



"""
Version-related utils.

Copyright 2024 Vlad Emelianov
"""

import contextlib
import importlib.metadata
from collections.abc import Iterable

from packaging.version import InvalidVersion, Version

from mypy_boto3_builder.constants import PACKAGE_NAME

VersionParts = tuple[int, ...]


def get_builder_version() -> str:
    """
    Get program version.
    """
    with contextlib.suppress(importlib.metadata.PackageNotFoundError):
        return importlib.metadata.version(PACKAGE_NAME)

    return "0.0.0"


def get_min_build_version(version: str) -> str:
    """
    Get min version build version by setting micro to 0.
    """
    major, minor, _ = Version(version).release
    return f"{major}.{minor}.0"


def get_release_version(version: str) -> str:
    """
    Get release version by removing post.
    """
    return Version(version).base_version


def get_max_build_version(version: str) -> str:
    """
    Get min version build version by bumping minor.
    """
    major, minor, _ = Version(version).release
    return f"{major}.{minor + 1}.0"


def bump_postrelease(version: str) -> str:
    """
    Bump postrelease version.
    """
    v = Version(version)
    major, minor, patch = v.release
    post = (v.post + 1) if v.post else 1
    return f"{major}.{minor}.{patch}.post{post}"


def sort_versions(version_list: Iterable[str]) -> list[str]:
    """
    Sort version list.
    """
    valid_versions = [i for i in version_list if is_valid_version(i)]
    valid_versions.sort(key=Version)
    return valid_versions


def is_valid_version(version: str) -> bool:
    """
    Check if version is valid.
    """
    try:
        Version(version)
    except InvalidVersion:
        return False
    return True


def stringify_parts(value: VersionParts) -> str:
    """
    Convert stable version tuple to string.
    """
    if not value:
        raise ValueError("Empty version parts")
    return ".".join(map(str, value))



"""
Type annotations for aioboto3.session module.

Copyright 2025 Vlad Emelianov
"""

from types import TracebackType
from typing import Any, Generic, TypeVar

from aioboto3.resources.base import AIOBoto3ServiceResource
from aioboto3.resources.factory import AIOBoto3ResourceFactory
from aiobotocore.config import AioConfig
from aiobotocore.credentials import AioCredentials
from aiobotocore.session import ClientCreatorContext
from boto3.session import Session as Boto3Session
from botocore.loaders import Loader
from botocore.session import Session as BotocoreSession

class Session(Boto3Session):
    def __init__(
        self,
        aws_access_key_id: str | None = ...,
        aws_secret_access_key: str | None = ...,
        aws_session_token: str | None = ...,
        region_name: str | None = ...,
        botocore_session: BotocoreSession | None = ...,
        profile_name: str | None = ...,
    ) -> None:
        self._session: BotocoreSession
        self.resource_factory: AIOBoto3ResourceFactory  # type: ignore[override]
        self._loader: Loader

    def get_credentials(self) -> AioCredentials | None: ...
    def client(  # type: ignore[override]
        self,
        service_name: str,
        region_name: str | None = ...,
        api_version: str | None = ...,
        use_ssl: bool | None = ...,
        verify: str | bool | None = ...,
        endpoint_url: str | None = ...,
        aws_access_key_id: str | None = ...,
        aws_secret_access_key: str | None = ...,
        aws_session_token: str | None = ...,
        config: AioConfig | None = ...,
    ) -> ClientCreatorContext[Any]: ...
    def resource(  # type: ignore[override]
        self,
        service_name: str,
        region_name: str | None = ...,
        api_version: str | None = ...,
        use_ssl: bool | None = ...,
        verify: bool | str | None = ...,
        endpoint_url: str | None = ...,
        aws_access_key_id: str | None = ...,
        aws_secret_access_key: str | None = ...,
        aws_session_token: str | None = ...,
        config: AioConfig | None = ...,
    ) -> ResourceCreatorContext[Any]: ...

_AIOBoto3ServiceResource = TypeVar("_AIOBoto3ServiceResource", bound=AIOBoto3ServiceResource)

class ResourceCreatorContext(Generic[_AIOBoto3ServiceResource]):
    def __init__(
        self,
        session: Session,
        service_name: str,
        region_name: str,
        api_version: str,
        use_ssl: bool,
        verify: bool,
        endpoint_url: str,
        aws_access_key_id: str,
        aws_secret_access_key: str,
        aws_session_token: str,
        config: AioConfig,
        resource_model: Any,
    ) -> None: ...
    async def __aenter__(self) -> _AIOBoto3ServiceResource: ...
    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        tb: TracebackType | None,
    ) -> None: ...



"""
Structure for mypy-boto3 package.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Iterable
from typing import TYPE_CHECKING

from mypy_boto3_builder.package_data import MypyBoto3PackageData
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.structures.package import Package
from mypy_boto3_builder.structures.packages.service_package import ServicePackage

if TYPE_CHECKING:
    from mypy_boto3_builder.type_annotations.type_literal import TypeLiteral


class MypyBoto3Package(Package):
    """
    Structure for mypy-boto3 package.

    Arguments:
        service_names -- List of included service names.
        service_packages -- List of included service packages.
        version -- Package version.
    """

    def __init__(
        self,
        service_names: Iterable[ServiceName],
        service_packages: Iterable[ServicePackage],
        version: str,
    ) -> None:
        super().__init__(MypyBoto3PackageData(), service_names, version=version)
        self.service_packages = list(service_packages)
        self.literals: list[TypeLiteral] = []



"""
Backport of `boto3.resources.model`.

Copyright 2024 Vlad Emelianov
"""

from dataclasses import dataclass
from typing import Any

from botocore import xform_name
from botocore.model import Shape, StructureShape


@dataclass
class Identifier:
    """
    A resource identifier, given by its name.
    """

    name: str
    member_name: str | None = None


class Action:
    """
    A service operation action.
    """

    def __init__(
        self,
        name: str,
        definition: dict[str, Any],
        resource_defs: dict[str, Any],
    ) -> None:
        self._definition = definition

        self.name = name
        self.request: Request | None = None
        if "request" in definition:
            self.request = Request(definition.get("request", {}))
        self.resource: ResponseResource | None = None
        if "resource" in definition:
            self.resource = ResponseResource(definition.get("resource", {}), resource_defs)
        self.path: str | None = definition.get("path")


class DefinitionWithParams:
    """
    An item which has parameters exposed via the ``params`` property.
    """

    def __init__(self, definition: dict[str, Any]) -> None:
        self._definition = definition

    @property
    def params(self) -> "list[Parameter]":
        """
        Get a list of auto-filled parameters for this request.
        """
        params: list[Parameter] = []

        params.extend(Parameter(**item) for item in self._definition.get("params", []))

        return params


class Parameter:
    """
    An auto-filled parameter which has a source and target.
    """

    def __init__(
        self,
        target: str,
        source: str,
        name: str | None = None,
        path: str | None = None,
        value: str | float | bool | None = None,  # noqa: FBT001
        **kwargs: str,
    ) -> None:
        self.target = target
        self.source = source
        self.name = name
        self.path = path
        self.value = value


class Request(DefinitionWithParams):
    """
    A service operation action request.
    """

    def __init__(self, definition: dict[str, Any]) -> None:
        super().__init__(definition)
        self.operation: str = definition.get("operation") or ""


class Waiter(DefinitionWithParams):
    """
    An event waiter specification.
    """

    PREFIX = "WaitUntil"

    def __init__(self, name: str, definition: dict[str, Any]) -> None:
        super().__init__(definition)
        self.name = name
        self.waiter_name: str | None = definition.get("waiterName")


class ResponseResource:
    """
    A resource response to create after performing an action.
    """

    def __init__(
        self, definition: dict[str, Any], resource_defs: dict[str, dict[str, Any]]
    ) -> None:
        self._definition = definition
        self._resource_defs = resource_defs
        self.type: str = definition.get("type") or ""
        self.path: str | None = definition.get("path")

    @property
    def identifiers(self) -> list[Parameter]:
        """
        A list of resource identifiers.
        """
        identifiers: list[Parameter] = []

        identifiers.extend(Parameter(**item) for item in self._definition.get("identifiers", []))

        return identifiers

    @property
    def model(self) -> "ResourceModel":
        """
        Get the resource model for the response resource.
        """
        return ResourceModel(self.type, self._resource_defs[self.type], self._resource_defs)


class Collection(Action):
    """
    A group of resources. See :py:class:`Action`.
    """

    @property
    def batch_actions(self) -> list[Action]:
        """
        Get a list of batch actions supported by the resource type.
        """
        if not self.resource:
            return []
        return self.resource.model.batch_actions


class ResourceModel:
    """
    A model representing a resource, defined via a JSON description.
    """

    def __init__(
        self, name: str, definition: dict[str, Any], resource_defs: dict[str, Any]
    ) -> None:
        self._definition = definition
        self._resource_defs = resource_defs
        self._renamed: dict[tuple[str, str], str] = {}

        self.name = name
        self.shape: str | None = definition.get("shape")

    def _get_name(self, category: str, name: str, snake_case: bool = True) -> str:  # noqa: FBT001, FBT002
        """
        Get a possibly renamed value given a category and name.
        """
        if snake_case:
            name = xform_name(name)

        return self._renamed.get((category, name), name)

    def get_attributes(self, shape: Shape) -> dict[str, tuple[str, Shape]]:
        """
        Get a dictionary of attribute names to original name and shape.
        """
        attributes: dict[str, tuple[str, Shape]] = {}
        identifier_names = [i.name for i in self.identifiers]
        if not isinstance(shape, StructureShape):
            return attributes
        for name, member in shape.members.items():
            snake_cased = xform_name(name)
            if snake_cased in identifier_names:
                # Skip identifiers, these are set through other means
                continue
            snake_cased = self._get_name("attribute", snake_cased, snake_case=False)
            attributes[snake_cased] = (name, member)

        return attributes

    @property
    def identifiers(self) -> list[Identifier]:
        """
        Get a list of resource identifiers.
        """
        identifiers: list[Identifier] = []

        for item in self._definition.get("identifiers", []):
            name = self._get_name("identifier", item["name"])
            member_name = item.get("memberName", None)
            if member_name:
                member_name = self._get_name("attribute", member_name)
            identifiers.append(Identifier(name, member_name))

        return identifiers

    @property
    def load(self) -> Action | None:
        """
        Get the load action for this resource, if it is defined.
        """
        definition = self._definition.get("load")

        if definition is None:
            return None

        return Action("load", definition, self._resource_defs)

    @property
    def actions(self) -> list[Action]:
        """
        Get a list of actions for this resource.
        """
        actions: list[Action] = []

        for name, item in self._definition.get("actions", {}).items():
            action_name = self._get_name("action", name)
            actions.append(Action(action_name, item, self._resource_defs))

        return actions

    @property
    def batch_actions(self) -> list[Action]:
        """
        Get a list of batch actions for this resource.
        """
        actions: list[Action] = []

        for name, item in self._definition.get("batchActions", {}).items():
            action_name = self._get_name("batch_action", name)
            actions.append(Action(action_name, item, self._resource_defs))

        return actions

    def _get_has_definition(self) -> dict[str, dict[str, Any]]:
        """
        Get a ``has`` relationship definition from a model.
        """
        if self.name not in self._resource_defs:
            # This is the service resource, so let us expose all of
            # the defined resources as subresources.
            definition: dict[str, dict[str, Any]] = {}

            for name, resource_def in self._resource_defs.items():
                # It's possible for the service to have renamed a
                # resource or to have defined multiple names that
                # point to the same resource type, so we need to
                # take that into account.
                found = False
                has_items = self._definition.get("has", {}).items()
                for has_name, has_def in has_items:
                    if has_def.get("resource", {}).get("type") == name:
                        definition[has_name] = has_def
                        found = True

                if not found:
                    fake_has: dict[str, dict[str, Any]] = {
                        "resource": {"type": name, "identifiers": []}
                    }

                    for identifier in resource_def.get("identifiers", []):
                        fake_has["resource"]["identifiers"].append(
                            {"target": identifier["name"], "source": "input"}
                        )

                    definition[name] = fake_has
        else:
            definition = self._definition.get("has", {})

        return definition

    def _get_related_resources(self, subresources: bool) -> list[Action]:  # noqa: FBT001
        """
        Get a list of sub-resources or references.
        """
        resources: list[Action] = []

        for name, definition in self._get_has_definition().items():
            if subresources:
                action_name = self._get_name("subresource", name, snake_case=False)
            else:
                action_name = self._get_name("reference", name)
            action = Action(action_name, definition, self._resource_defs)
            if not action.resource:
                continue

            data_required = False
            for identifier in action.resource.identifiers:
                if identifier.source == "data":
                    data_required = True
                    break

            if (subresources and not data_required) or (not subresources and data_required):
                resources.append(action)

        return resources

    @property
    def subresources(self) -> list[Action]:
        """
        Get a list of sub-resources.
        """
        return self._get_related_resources(subresources=True)

    @property
    def references(self) -> list[Action]:
        """
        Get a list of reference resources.
        """
        return self._get_related_resources(subresources=False)

    @property
    def collections(self) -> list[Collection]:
        """
        Get a list of collections for this resource.
        """
        collections: list[Collection] = []

        for name, item in self._definition.get("hasMany", {}).items():
            collection_name = self._get_name("collection", name)
            collections.append(Collection(collection_name, item, self._resource_defs))

        return collections

    @property
    def waiters(self) -> list[Waiter]:
        """
        Get a list of waiters for this resource.
        """
        waiters: list[Waiter] = []

        for name, item in self._definition.get("waiters", {}).items():
            waiter_name = self._get_name("waiter", Waiter.PREFIX + name)
            waiters.append(Waiter(waiter_name, item))

        return waiters



"""
Writer for package static and template files.

Copyright 2024 Vlad Emelianov
"""

import filecmp
import shutil
from collections.abc import Iterable, Sequence
from pathlib import Path
from typing import Final

from mypy_boto3_builder.constants import TEMPLATES_PATH
from mypy_boto3_builder.enums.service_module_name import ServiceModuleName
from mypy_boto3_builder.logger import get_logger
from mypy_boto3_builder.structures.package import Package
from mypy_boto3_builder.structures.packages.service_package import ServicePackage
from mypy_boto3_builder.utils.markdown import fix_pypi_headers
from mypy_boto3_builder.utils.path import print_path, walk_path
from mypy_boto3_builder.writers.ruff_formatter import RuffFormatter
from mypy_boto3_builder.writers.utils import (
    format_md,
    insert_md_toc,
    render_jinja2_package_template,
)


class TemplateRender:
    """
    Template render target.
    """

    def __init__(
        self,
        template_path: Path,
        path: Path | None = None,
        paths: tuple[Path, ...] = (),
    ) -> None:
        self.template_path = template_path
        self.paths = (*([path] if path else []), *paths)


class PackageWriter:
    """
    Writer for package static and template files.

    Arguments:
        output_path -- Output path
        generate_package -- Whether to generate setup files
        cleanup -- Whether to remove unknown files
        is_typings -- Whether to generate typings without `-stubs` suffix
    """

    _PY_EXTENSIONS: Final = {".py", ".pyi"}
    _MD_EXTENSIONS: Final = {".md"}

    def __init__(
        self,
        output_path: Path,
        *,
        generate_package: bool,
        cleanup: bool,
        is_typings: bool = True,
    ) -> None:
        self.output_path = output_path
        self.is_package = generate_package
        self.is_typings = is_typings
        self.cleanup = cleanup
        self.logger = get_logger()

    def _get_package_path(self, package: Package) -> Path:
        if self.is_package:
            return self.output_path / package.directory_name / package.name

        if self.is_typings:
            return self.output_path / package.library_name

        return self.output_path / package.name

    def _get_service_package_path(self, package: ServicePackage) -> Path:
        if self.is_package:
            return self.output_path / package.directory_name / package.name

        return self.output_path / package.name

    def _get_setup_path(self, package: Package) -> Path:
        return self.output_path / package.directory_name

    def _write_static_paths(
        self,
        static_files_path: Path | None,
        package: Package,
        exclude_paths: Iterable[Path],
    ) -> list[Path]:
        if not static_files_path:
            return []
        package_path = self._get_package_path(package)
        result: list[Path] = []
        for static_path in static_files_path.glob("**/*.pyi"):
            relative_output_path = static_path.relative_to(static_files_path)
            file_path = package_path / relative_output_path
            if file_path in exclude_paths:
                continue
            result.append(file_path)
            file_path.parent.mkdir(exist_ok=True, parents=True)
            if not file_path.exists():
                shutil.copy(static_path, file_path)
                self.logger.debug(f"Created {print_path(file_path)}")
            elif not filecmp.cmp(static_path, file_path):
                shutil.copy(static_path, file_path)
                self.logger.debug(f"Updated {print_path(file_path)}")
        return result

    def _get_setup_template_paths(
        self,
        package: Package,
        templates_path: Path | None,
    ) -> list[TemplateRender]:
        if not templates_path or not self.is_package:
            return []

        result: list[TemplateRender] = []
        setup_path = self._get_setup_path(package)
        template_paths = (
            templates_path / "setup.py.jinja2",
            templates_path / "README.md.jinja2",
            TEMPLATES_PATH / "common" / "LICENSE.jinja2",
        )
        for template_path in template_paths:
            file_name = template_path.stem
            output_file_path = setup_path / file_name
            result.append(TemplateRender(template_path, output_file_path))
        return result

    def _get_package_template_renders(
        self,
        package: Package,
        templates_path: Path | None,
    ) -> list[TemplateRender]:
        if not templates_path or not package.has_main_package():
            return []

        result: list[TemplateRender] = []
        package_path = self._get_package_path(package)
        package_template_path = templates_path / package.name
        for template_path in walk_path(package_template_path, glob_pattern="**/*.jinja2"):
            file_name = template_path.stem
            output_file_path = (
                package_path / template_path.relative_to(package_template_path).parent / file_name
            )
            result.append(TemplateRender(template_path, output_file_path))
        return result

    def _render_template(
        self,
        template_path: Path,
        render_paths: Iterable[Path],
        package: Package,
    ) -> None:
        content = render_jinja2_package_template(template_path, package=package)
        for output_path in render_paths:
            file_suffix = output_path.suffix.lower()
            if file_suffix in self._MD_EXTENSIONS:
                content = insert_md_toc(content)
                content = fix_pypi_headers(content)
                content = format_md(content)
            self._write_template(output_path, content)

    def _render_templates(
        self,
        package: Package,
        template_renders: Iterable[TemplateRender],
    ) -> None:
        for template_render in template_renders:
            self._render_template(template_render.template_path, template_render.paths, package)

    def _write_template(self, path: Path, content: str) -> None:
        if not path.parent.exists():
            path.parent.mkdir(exist_ok=True, parents=True)

        path.write_text(content)
        self.logger.debug(f"Rendered {print_path(path)}", tags=print_path(path))

    def _render_docs_templates(
        self,
        package: Package,
        template_renders: Iterable[TemplateRender],
    ) -> None:
        for template_render in template_renders:
            content = render_jinja2_package_template(template_render.template_path, package=package)
            for output_path in template_render.paths:
                self._write_template(output_path, content)

    def _cleanup(self, valid_paths: Sequence[Path], output_path: Path) -> None:
        if not self.cleanup or not output_path.exists():
            return
        for unknown_path in walk_path(output_path, valid_paths):
            unknown_path.unlink()
            self.logger.debug(f"Deleted {print_path(unknown_path)}", tags=print_path(unknown_path))

    def write_package(
        self,
        package: Package,
        template_path: Path | None = None,
        static_files_path: Path | None = None,
        include_template_names: Sequence[str] = (),
        exclude_template_names: Sequence[str] = (),
    ) -> None:
        """
        Generate files for a package.

        Arguments:
            package -- Package to render
            template_path -- Path to Jinja templates for package
            static_files_path -- Path to static files for package
            include_template_names -- Render only templates with these names
            exclude_template_names -- Do not render templates with these names
        """
        self.logger.debug(
            f"Writing {package.data.pypi_name} to {print_path(self.output_path)}",
            tags=(package.data.pypi_name, print_path(self.output_path)),
        )
        template_renders: list[TemplateRender] = [
            *self._get_setup_template_paths(package, template_path),
            *self._get_package_template_renders(package, template_path),
        ]
        if include_template_names:
            template_renders = list(
                filter(lambda x: x.template_path.name in include_template_names, template_renders)
            )
        if exclude_template_names:
            template_renders = list(
                filter(
                    lambda x: x.template_path.name not in exclude_template_names, template_renders
                )
            )
        exclude_static_paths: set[Path] = set()
        for template_render in template_renders:
            exclude_static_paths.update(template_render.paths)
        static_paths = self._write_static_paths(static_files_path, package, exclude_static_paths)
        self._render_templates(package, template_renders)

        rendered_paths = [path for t in template_renders for path in t.paths]
        valid_paths = (*rendered_paths, *static_paths)

        self._format_output(package, valid_paths)

        cleanup_path = self._get_cleanup_path(package)
        if cleanup_path:
            self._cleanup(valid_paths, cleanup_path)

    def _get_cleanup_path(self, package: Package) -> Path | None:
        if self.is_package:
            return self._get_setup_path(package)

        if package.has_main_package():
            return self._get_package_path(package)

        return None

    def _format_output(self, package: Package, paths: Sequence[Path]) -> None:
        ruff_formatter = RuffFormatter(
            known_first_party=[package.name] if package.has_main_package() else [],
            known_third_party=[
                "boto3",
                "botocore",
                "aioboto3",
                "aiobotocore",
                *[package.data.get_service_package_name(i) for i in package.service_names],
            ],
        )
        existing_paths = [path for path in paths if path.exists()]
        format_python_paths = [
            path for path in existing_paths if path.suffix.lower() in self._PY_EXTENSIONS
        ]
        if format_python_paths:
            ruff_formatter.format_python(format_python_paths)

        format_md_paths = [
            path for path in existing_paths if path.suffix.lower() in self._MD_EXTENSIONS
        ]
        if format_md_paths:
            for path in format_md_paths:
                content = path.read_text()
                content = ruff_formatter.format_markdown(content)
                path.write_text(content)

    def write_docs(self, package: Package, templates_path: Path) -> None:
        """
        Generate docs for a package.
        """
        self.output_path.mkdir(exist_ok=True, parents=True)
        template_renders: list[TemplateRender] = []
        for template_path in templates_path.glob("**/*.jinja2"):
            file_name = template_path.stem
            template_renders.append(TemplateRender(template_path, self.output_path / file_name))

        self._render_docs_templates(package, template_renders)

    def _get_service_package_template_paths(
        self,
        package: ServicePackage,
        templates_path: Path,
    ) -> list[TemplateRender]:
        module_templates_path = templates_path / "service"
        package_path = self._get_service_package_path(package)
        file_paths: list[TemplateRender] = [
            TemplateRender(
                module_templates_path / "version.py.jinja2",
                package_path / "version.py",
            ),
            TemplateRender(
                module_templates_path / "__init__.pyi.jinja2",
                paths=(
                    package_path / "__init__.pyi",
                    package_path / "__init__.py",
                ),
            ),
            TemplateRender(
                module_templates_path / "__main__.py.jinja2",
                package_path / "__main__.py",
            ),
            TemplateRender(module_templates_path / "py.typed.jinja2", package_path / "py.typed"),
            TemplateRender(
                module_templates_path / ServiceModuleName.client.template_name,
                paths=(
                    package_path / ServiceModuleName.client.stub_file_name,
                    package_path / ServiceModuleName.client.file_name,
                ),
            ),
        ]
        if package.service_resource:
            file_paths.append(
                TemplateRender(
                    module_templates_path / ServiceModuleName.service_resource.template_name,
                    paths=(
                        package_path / ServiceModuleName.service_resource.stub_file_name,
                        package_path / ServiceModuleName.service_resource.file_name,
                    ),
                ),
            )
        if package.paginators:
            file_paths.append(
                TemplateRender(
                    module_templates_path / ServiceModuleName.paginator.template_name,
                    paths=(
                        package_path / ServiceModuleName.paginator.stub_file_name,
                        package_path / ServiceModuleName.paginator.file_name,
                    ),
                ),
            )
        if package.waiters:
            file_paths.append(
                TemplateRender(
                    module_templates_path / ServiceModuleName.waiter.template_name,
                    paths=(
                        package_path / ServiceModuleName.waiter.stub_file_name,
                        package_path / ServiceModuleName.waiter.file_name,
                    ),
                ),
            )
        if package.literals:
            file_paths.append(
                TemplateRender(
                    module_templates_path / ServiceModuleName.literals.template_name,
                    paths=(
                        package_path / ServiceModuleName.literals.stub_file_name,
                        package_path / ServiceModuleName.literals.file_name,
                    ),
                ),
            )
        if package.type_defs:
            file_paths.append(
                TemplateRender(
                    module_templates_path / ServiceModuleName.type_defs.template_name,
                    paths=(
                        package_path / ServiceModuleName.type_defs.stub_file_name,
                        package_path / ServiceModuleName.type_defs.file_name,
                    ),
                ),
            )
        return file_paths

    def write_service_package(self, package: ServicePackage, templates_path: Path) -> None:
        """
        Create stubs files for service.

        Arguments:
            package -- Service package.
        """
        template_renders: list[TemplateRender] = [
            *self._get_setup_template_paths(package, templates_path),
            *self._get_service_package_template_paths(package, templates_path),
        ]
        self._render_templates(package, template_renders)

        valid_paths = [path for t in template_renders for path in t.paths]

        self._format_output(package, valid_paths)

        output_path = (
            self._get_setup_path(package)
            if self.is_package
            else self._get_service_package_path(package)
        )
        self._cleanup(valid_paths, output_path)

    def write_service_docs(self, package: ServicePackage, templates_path: Path) -> None:
        """
        Create service docs files.

        Arguments:
            package -- Service package.
            output_path -- Path to output directory.
        """
        docs_path = self.output_path / package.name
        docs_path.mkdir(exist_ok=True, parents=True)
        template_renders: list[TemplateRender] = [
            TemplateRender(templates_path / "README.md.jinja2", docs_path / "README.md"),
            TemplateRender(templates_path / "client.md.jinja2", docs_path / "client.md"),
            TemplateRender(templates_path / "usage.md.jinja2", docs_path / "usage.md"),
        ]

        if package.literals:
            template_renders.append(
                TemplateRender(templates_path / "literals.md.jinja2", docs_path / "literals.md"),
            )

        if package.type_defs:
            template_renders.append(
                TemplateRender(templates_path / "type_defs.md.jinja2", docs_path / "type_defs.md"),
            )

        if package.waiters:
            template_renders.append(
                TemplateRender(templates_path / "waiters.md.jinja2", docs_path / "waiters.md"),
            )

        if package.paginators:
            template_renders.append(
                TemplateRender(
                    templates_path / "paginators.md.jinja2",
                    docs_path / "paginators.md",
                ),
            )

        if package.service_resource:
            template_renders.append(
                TemplateRender(
                    templates_path / "service_resource.md.jinja2",
                    docs_path / "service_resource.md",
                ),
            )

        self._render_docs_templates(package, template_renders)
        valid_paths = [path for t in template_renders for path in t.paths]
        self._cleanup(valid_paths, docs_path)



"""
Type annotations for aiobotocore.regions module.

Copyright 2025 Vlad Emelianov
"""

from logging import Logger
from typing import Any, Mapping

from botocore.endpoint_provider import RuleSetEndpoint
from botocore.model import OperationModel
from botocore.regions import EndpointRulesetResolver

LOG: Logger

class AioEndpointRulesetResolver(EndpointRulesetResolver):
    async def construct_endpoint(  # type: ignore[override]
        self,
        operation_model: OperationModel,
        call_args: Mapping[str, Any] | None,
        request_context: Mapping[str, Any] | None,
    ) -> RuleSetEndpoint: ...



"""
Exceptions for mypy_boto3_builder.

Copyright 2024 Vlad Emelianov
"""


class BuilderError(ValueError):
    """
    Base error for mypy_boto3_builder.
    """


class JinjaManagerError(BuilderError):
    """
    Base JinjaManager exception.
    """


class ShapeParserError(BuilderError):
    """
    Main error for ShapeParser.
    """


class BuildError(BuilderError):
    """
    Error during build process.
    """


class BuildEnvError(BuildError):
    """
    Error during getting 3rd party data.
    """


class BuildInternalError(BuildError):
    """
    Error on invalid build.
    """


class TypeAnnotationError(BuildInternalError):
    """
    Error on invalid internal type annotation.
    """


class StructureError(BuildInternalError):
    """
    Error on invalid internal structure.
    """


class AlreadyPublishedError(BuildInternalError):
    """
    Error on already published package.
    """



"""
Boto3 ServiceResource sub-Resource.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Generator
from typing import TYPE_CHECKING

from mypy_boto3_builder.import_helpers.import_helper import Import
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.structures.class_record import ClassRecord
from mypy_boto3_builder.type_annotations.external_import import ExternalImport
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation

if TYPE_CHECKING:
    from mypy_boto3_builder.structures.collection import Collection


class ResourceRecord(ClassRecord):
    """
    Boto3 ServiceResource sub-Resource.
    """

    def __init__(self, name: str, service_name: ServiceName) -> None:
        super().__init__(
            name=name,
            bases=[ExternalImport(Import.boto3 + "resources" + "base", "ServiceResource")],
            use_alias=True,
        )
        self.service_name: ServiceName = service_name
        self.collections: list[Collection] = []

    @property
    def boto3_doc_link(self) -> str:
        """
        Link to boto3 docs.
        """
        return (
            f"{self.service_name.boto3_doc_link_parent}/{self.name.lower()}/index.html"
            f"#{self.service_name.class_name}.{self.name}"
        )

    def iterate_types(self) -> Generator[FakeAnnotation]:
        """
        Iterate over all type annotations from collections.
        """
        yield from super().iterate_types()
        for collection in self.collections:
            yield from collection.iterate_types()



"""
Wrapper for constant like `False` or `"test"`.

Copyright 2024 Vlad Emelianov
"""

from typing import Final, Self

from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation


class EllipsisType:
    """
    Placeholder for `...`.
    """


ValueType = str | int | float | EllipsisType | None


class TypeConstant(FakeAnnotation):
    """
    Wrapper for constant like `False` or `"test"`.

    Arguments:
        value -- Constant value.
    """

    Ellipsis: Final[EllipsisType] = EllipsisType()

    def __init__(self, value: ValueType) -> None:
        self.value: ValueType = value

    def render(self) -> str:
        """
        Render type annotation to a valid Python code for local usage.

        Returns:
            A string with a valid type annotation.
        """
        if self.value is self.Ellipsis:
            return "..."

        return repr(self.value)

    def __copy__(self) -> Self:
        """
        Create a copy of type annotation wrapper.
        """
        return self.__class__(self.value)



"""
Description for boto3 service.

Copyright 2024 Vlad Emelianov
"""

from typing import ClassVar, Final

from mypy_boto3_builder.constants import ALL
from mypy_boto3_builder.utils.strings import is_reserved

__all__ = (
    "ServiceName",
    "ServiceNameCatalog",
)


class ServiceName:
    """
    Description for boto3 service.
    """

    ALL: Final = "all"
    UPDATED: Final = "updated"
    ESSENTIAL: Final = "essential"
    FULL: Final = "full"
    LATEST: Final = "latest"

    ESSENTIAL_NAMES: Final = {
        "ec2",
        "rds",
        "s3",
        "lambda",
        "sqs",
        "cloudformation",
        "dynamodb",
    }
    CONDA_FORGE_AVAILABLE: Final = {
        "ec2",
        "rds",
        "s3",
        "lambda",
        "sqs",
        "cloudformation",
        "dynamodb",
    }

    def __init__(self, name: str, class_name: str, override_boto3_name: str = "") -> None:
        self.name = name
        self.class_name = class_name
        self.boto3_version = self.LATEST
        self.override_boto3_name = override_boto3_name

    def __hash__(self) -> int:
        """
        Calculate hash value based on service name.
        """
        return hash(self.name)

    def __str__(self) -> str:
        """
        Represent as string for debugging.
        """
        return f"<ServiceName {self.name} {self.class_name}>"

    @property
    def underscore_name(self) -> str:
        """
        Python-friendly service name.
        """
        return self.name.replace("-", "_")

    @property
    def boto3_name(self) -> str:
        """
        Boto3 package name.
        """
        return self.override_boto3_name or self.name

    @property
    def import_name(self) -> str:
        """
        Safe mudule import name.
        """
        name = self.name.replace("-", "_")
        if is_reserved(name):
            return f"{name}_"

        return name

    @property
    def extras_name(self) -> str:
        """
        Extras name for subpackage installation.
        """
        return self.name

    def is_essential(self) -> bool:
        """
        Whether service is included to `types-library[essential]`.
        """
        return self.name in self.ESSENTIAL_NAMES

    def is_conda_forge_available(self) -> bool:
        """
        Whether service is available for `conda-forge`.
        """
        return self.name in self.CONDA_FORGE_AVAILABLE

    @property
    def boto3_doc_link(self) -> str:
        """
        Link to boto3 docs.
        """
        return (
            "https://boto3.amazonaws.com/v1/documentation/api/"
            f"latest/reference/services/{self.boto3_name}.html#{self.class_name.lower()}"
        )

    @property
    def boto3_doc_link_parent(self) -> str:
        """
        Link to boto3 docs parent directory.
        """
        return (
            "https://boto3.amazonaws.com/v1/documentation/api/"
            f"latest/reference/services/{self.boto3_name}"
        )

    def get_client_name(self) -> str:
        """
        Get Client name.
        """
        return f"{self.class_name}Client"

    def get_service_resource_name(self) -> str:
        """
        Get ServiceResource name.
        """
        return f"{self.class_name}ServiceResource"


class ServiceNameCatalog:
    """
    Finder for botocore services by name.
    """

    all = ServiceName("__all", "__all")
    ec2 = ServiceName("ec2", "EC2")
    iam = ServiceName("iam", "IAM")
    s3 = ServiceName("s3", "S3")
    rds = ServiceName("rds", "RDS")
    cloudwatch = ServiceName("cloudwatch", "CloudWatch")
    opsworks = ServiceName("opsworks", "OpsWorks")
    sns = ServiceName("sns", "SNS")
    glacier = ServiceName("glacier", "Glacier")
    dynamodb = ServiceName("dynamodb", "DynamoDB")
    sqs = ServiceName("sqs", "SQS")
    cloudformation = ServiceName("cloudformation", "CloudFormation")
    cloudsearchdomain = ServiceName("cloudsearchdomain", "CloudSearchDomain")
    logs = ServiceName("logs", "CloudWatchLogs")
    lambda_ = ServiceName("lambda", "Lambda")
    stepfunctions = ServiceName("stepfunctions", "SFN")
    old_redshift_serverless = ServiceName(
        "redshift-serverless",
        "RedshiftServerless",
        "redshiftserverless",
    )
    old_ssm_sap = ServiceName("ssm-sap", "SsmSap", "ssmsap")
    sms_voice = ServiceName("sms-voice", "SMSVoice")

    ITEMS: ClassVar[dict[str, ServiceName]] = {
        ec2.boto3_name: ec2,
        iam.boto3_name: iam,
        s3.boto3_name: s3,
        rds.boto3_name: rds,
        cloudwatch.boto3_name: cloudwatch,
        opsworks.boto3_name: opsworks,
        sns.boto3_name: sns,
        glacier.boto3_name: glacier,
        dynamodb.boto3_name: dynamodb,
        sqs.boto3_name: sqs,
        cloudformation.boto3_name: cloudformation,
        cloudsearchdomain.boto3_name: cloudsearchdomain,
        logs.boto3_name: logs,
        lambda_.boto3_name: lambda_,
        old_redshift_serverless.boto3_name: old_redshift_serverless,
        old_ssm_sap.boto3_name: old_ssm_sap,
        sms_voice.boto3_name: sms_voice,
    }

    @classmethod
    def add(cls, name: str, class_name: str) -> ServiceName:
        """
        Add new ServiceName to catalog or modify existing one.

        Returns:
            New ServiceName or modified if it exists.
        """
        if name in cls.ITEMS:
            service_name = cls.ITEMS[name]
            service_name.class_name = class_name
            return service_name

        service_name = ServiceName(name, class_name)
        cls.ITEMS[name] = service_name
        return service_name

    @classmethod
    def to_str(cls, service_name: ServiceName) -> str:
        """
        Represent ServiceName as string for lookups.
        """
        if service_name is cls.all:
            return ALL
        return service_name.name



"""
Base class for ClassRecord.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Generator

from mypy_boto3_builder.import_helpers.import_helper import Import
from mypy_boto3_builder.type_annotations.external_import import ExternalImport
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.utils.type_checks import is_type_subscript


class BaseClass:
    """
    Base class for ClassRecord.
    """

    def __init__(self, name: str, annotation: FakeAnnotation) -> None:
        self.name = name
        self.annotation = annotation

    def render(self) -> str:
        """
        Render for usage as class base.

        If annotation is a subscript, return name.
        """
        if is_type_subscript(self.annotation):
            return self.name

        return self.annotation.render()

    def render_definition(self) -> str:
        """
        Render definition that works in runtime and TYPE_CHECKING.

        Returns empty string if annotation is not a subscript.
        """
        annotation = self.annotation
        if not is_type_subscript(annotation):
            return ""

        return (
            "if TYPE_CHECKING:\n"
            f"    {self.name} = {annotation.render_definition()}\n"
            "else:\n"
            f"    {self.name} = {annotation.parent.render_definition()}"
            "  # type: ignore[assignment]\n"
        )

    def iterate_types(self) -> Generator[FakeAnnotation]:
        """
        Iterate over definition type annotations.
        """
        yield from self.annotation.iterate_types()
        if is_type_subscript(self.annotation):
            yield ExternalImport(Import.typing, "TYPE_CHECKING")



"""
Type annotations for aiobotocore.config module.

Copyright 2025 Vlad Emelianov
"""

from typing import Any, TypeVar

from aiobotocore.httpsession import AIOHTTPSession
from botocore.config import Config

_Config = TypeVar("_Config", bound=Config)

class AioConfig(Config):
    def __init__(
        self, connector_args: Any = ..., http_session_cls: type[AIOHTTPSession] = ..., **kwargs: Any
    ) -> None: ...
    def merge(self: _Config, other_config: _Config) -> _Config: ...



"""
String to type annotation map to replace overriden botocore literals.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Mapping
from typing import Final

from mypy_boto3_builder.service_name import ServiceName, ServiceNameCatalog
from mypy_boto3_builder.type_annotations.type_literal import TypeLiteral
from mypy_boto3_builder.utils.lookup_dict import LookupDict

# Mapping overriding literal types for botocore literals.
# ServiceName -> Literal name -> Literal children
# Missing value is parsed from botocore shapes.
LITERAL_TYPE_MAP: Final[Mapping[ServiceName, Mapping[str, set[str]]]] = {
    # FIXME: https://github.com/boto/botocore/issues/3128
    ServiceNameCatalog.ec2: {"PlatformValuesType": {"windows"}},
}

_LOOKUP: LookupDict[set[str]] = LookupDict(
    {ServiceNameCatalog.to_str(k): v for k, v in LITERAL_TYPE_MAP.items()},
)


def get_type_literal_stub(service_name: ServiceName, literal_name: str) -> TypeLiteral | None:
    """
    Get stub type for botocore literal.

    Arguments:
        service_name -- Service name.
        literal_name -- Target Literal name.

    Returns:
        Literal children or None.
    """
    literal_children = _LOOKUP.get(service_name.name, literal_name)
    if not literal_children:
        return None

    return TypeLiteral(literal_name, literal_children)



"""
Parser for botocore ServiceResource, produces `structures.ServiceResource`.

Copyright 2024 Vlad Emelianov
"""

from mypy_boto3_builder.boto3_ports.model import ResourceModel
from mypy_boto3_builder.constants import SERVICE_RESOURCE
from mypy_boto3_builder.enums.service_module_name import ServiceModuleName
from mypy_boto3_builder.parsers.resource_parser import ResourceParser
from mypy_boto3_builder.parsers.shape_parser import ShapeParser
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.structures.attribute import Attribute
from mypy_boto3_builder.structures.method import Method
from mypy_boto3_builder.structures.resource_record import ResourceRecord
from mypy_boto3_builder.structures.service_resource import ServiceResource
from mypy_boto3_builder.type_annotations.internal_import import InternalImport
from mypy_boto3_builder.type_maps.service_stub_map import get_stub_method_map
from mypy_boto3_builder.utils.strings import get_class_prefix


class ServiceResourceParser(ResourceParser):
    """
    Parser for botocore ServiceResource, produces `structures.ServiceResource`.

    Arguments:
        service_name -- Target service name.
        shape_parser -- ShapeParser instance.
    """

    def __init__(
        self,
        service_name: ServiceName,
        shape_parser: ShapeParser,
    ) -> None:
        super().__init__(
            name=SERVICE_RESOURCE,
            service_name=service_name,
            shape_parser=shape_parser,
        )
        self.alias = self.service_name.get_service_resource_name()

    def _get_meta_attribute(self, name: str) -> Attribute:
        return Attribute(
            "meta",
            InternalImport(
                name,
                self.service_name,
                ServiceModuleName.service_resource,
            ),
            type_ignore="override",
        )

    def _parse_method_map(self) -> dict[str, Method]:
        shape_method_map = self.shape_parser.get_service_resource_method_map()
        stub_method_map = get_stub_method_map(self.service_name, self.name)
        for method in stub_method_map.values():
            self.shape_parser.create_request_type_annotation(
                method=method,
                name=f"{self.name}{get_class_prefix(method.name)}Request",
            )
        return {**shape_method_map, **stub_method_map}

    def _get_resource_model(self) -> ResourceModel:
        return self.shape_parser.get_service_resource_model()

    def parse_service_resource(self) -> ServiceResource | None:
        """
        Parse main botocore ServiceResource.
        """
        if not self.shape_parser.has_service_resource():
            return None

        self._logger.debug(f"Parsing {self.alias}", tags=self.alias)
        result = ServiceResource(
            name=self.alias,
            service_name=self.service_name,
        )
        meta_attribute = self._get_meta_attribute(result.resource_meta_class.name)

        self._logger.debug(f"Parsing {self.alias} methods", tags=self.alias)
        method_map = self._parse_method_map()
        result.methods.extend(list(method_map.values()))

        self._logger.debug(f"Parsing {self.alias} attributes and collections", tags=self.alias)
        collections = self._parse_collections()
        result.attributes.append(meta_attribute)
        result.attributes.extend(self._parse_attributes(collections))
        result.collections.extend(collections)

        sub_resources = self._parse_sub_resources(meta_attribute)
        result.sub_resources.extend(sub_resources)

        return result

    def _parse_sub_resources(self, meta_attribute: Attribute) -> list[ResourceRecord]:
        result: list[ResourceRecord] = []
        for sub_resource_name in self.shape_parser.get_subresource_names():
            self._logger.debug(
                f"Parsing {sub_resource_name} sub resource",
                tags=sub_resource_name,
            )
            resource_parser = ResourceParser(
                service_name=self.service_name,
                name=sub_resource_name,
                shape_parser=self.shape_parser,
            )
            sub_resource = resource_parser.parse()
            sub_resource.attributes.append(meta_attribute.copy())
            result.append(sub_resource)
        return result



"""
Wrapper for simple type annotations from this module.

Copyright 2024 Vlad Emelianov
"""

from typing import Self

from mypy_boto3_builder.enums.service_module_name import ServiceModuleName
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation


class InternalImport(FakeAnnotation):
    """
    Wrapper for simple type annotations from this module.

    Arguments:
        name -- Import name.
        service_name -- Service that import belongs to.
        module_name -- Service module name.
        use_alias -- Use name alias.
    """

    def __init__(
        self,
        name: str,
        service_name: ServiceName | None = None,
        module_name: ServiceModuleName = ServiceModuleName.service_resource,
        *,
        use_alias: bool = False,
    ) -> None:
        self.name: str = name
        self.service_name: ServiceName | None = service_name
        self.module_name: ServiceModuleName = module_name
        self.use_alias: bool = use_alias

    @property
    def alias_name(self) -> str:
        """
        Import name alias.
        """
        return f"_{self.name}"

    def render(self) -> str:
        """
        Render type annotation to a valid Python code for local usage.

        Returns:
            A string with a valid type annotation.
        """
        if self.use_alias:
            return self.alias_name

        return self.name

    def __copy__(self) -> Self:
        """
        Create a copy of type annotation wrapper.
        """
        return self.__class__(
            self.name,
            self.service_name,
            self.module_name,
            use_alias=self.use_alias,
        )



"""
Parser for botocore ServiceResource sub-resource, produces `structures.Resource`.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Iterable

from mypy_boto3_builder.boto3_ports.model import ResourceModel
from mypy_boto3_builder.logger import get_logger
from mypy_boto3_builder.parsers.parse_attributes import parse_attributes
from mypy_boto3_builder.parsers.parse_collections import parse_collections
from mypy_boto3_builder.parsers.parse_references import parse_references
from mypy_boto3_builder.parsers.shape_parser import ShapeParser
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.structures.attribute import Attribute
from mypy_boto3_builder.structures.collection import Collection
from mypy_boto3_builder.structures.method import Method
from mypy_boto3_builder.structures.resource_record import ResourceRecord
from mypy_boto3_builder.type_annotations.internal_import import InternalImport
from mypy_boto3_builder.type_maps.service_stub_map import get_stub_method_map
from mypy_boto3_builder.utils.strings import get_class_prefix


class ResourceParser:
    """
    Parser for botocore ServiceResource sub-resource, produces `structures.Resource`.

    Arguments:
        name -- Resource name.
        service_name -- Target service name.
        shape_parser -- ShapeParser instance.
    """

    def __init__(
        self,
        name: str,
        service_name: ServiceName,
        shape_parser: ShapeParser,
    ) -> None:
        self.name = name
        self.service_name = service_name
        self.shape_parser = shape_parser
        self._resource_model: ResourceModel | None = None
        self._logger = get_logger()

    def _get_resource_model(self) -> ResourceModel:
        return self.shape_parser.get_resource_model(self.name)

    @property
    def resource_model(self) -> ResourceModel:
        """
        Resource model.
        """
        if not self._resource_model:
            self._resource_model = self._get_resource_model()
        return self._resource_model

    def _parse_method_map(self) -> dict[str, Method]:
        shape_method_map = self.shape_parser.get_resource_method_map(self.name)
        stub_method_map = get_stub_method_map(self.service_name, self.name)
        for method in stub_method_map.values():
            self.shape_parser.create_request_type_annotation(
                method=method,
                name=f"{self.name}{get_class_prefix(method.name)}Request",
            )
        return {**shape_method_map, **stub_method_map}

    def _parse_attributes(self, collections: Iterable[Collection]) -> list[Attribute]:
        identifier_attributes = self.shape_parser.get_identifier_attributes(self.name)
        references = parse_references(self.resource_model)
        attributes = parse_attributes(
            self.service_name,
            self.name,
            self.resource_model,
            self.shape_parser,
        )

        existing_attribute_names = {a.name for a in identifier_attributes}
        for attribute in references:
            if attribute.name in existing_attribute_names:
                self._logger.warning(
                    f"Duplicate attribute name {self.name}.{attribute.name} in reference, renaming"
                )
                attribute.name = f"{attribute.name}_reference"
            existing_attribute_names.add(attribute.name)

        collection_attributes: list[Attribute] = []
        for collection in collections:
            attribute = Attribute(
                collection.attribute_name,
                InternalImport(collection.name, self.service_name),
                is_collection=True,
            )
            if attribute.name in existing_attribute_names:
                self._logger.warning(
                    f"Duplicate attribute name {self.name}.{attribute.name} in collection, renaming"
                )
                attribute.name = f"{attribute.name}_collection"
            collection_attributes.append(attribute)
            existing_attribute_names.add(attribute.name)

        for attribute in attributes:
            if attribute.name in existing_attribute_names:
                # ResourceModel.load_rename_map logic
                attribute.name = f"{attribute.name}_attribute"
            existing_attribute_names.add(attribute.name)

        return [
            *identifier_attributes,
            *references,
            *collection_attributes,
            *attributes,
        ]

    def _parse_collections(self) -> list[Collection]:
        return parse_collections(
            self.service_name, self.name, self.resource_model, self.shape_parser
        )

    def parse(self) -> ResourceRecord:
        """
        Parse botocore sub Resource data.

        Returns:
            Resource structure.
        """
        result = ResourceRecord(
            name=self.name,
            service_name=self.service_name,
        )

        method_map = self._parse_method_map()
        if self.resource_model.load:
            method_map["load"] = self.shape_parser.get_resource_load_method("load")
            method_map["reload"] = self.shape_parser.get_resource_load_method("reload")

        result.methods.extend(list(method_map.values()))
        collections = self._parse_collections()
        result.attributes.extend(self._parse_attributes(collections))
        result.collections.extend(collections)

        return result



"""
Parser for botocore shape files.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Iterable, Sequence
from typing import TYPE_CHECKING

from botocore.eventstream import EventStream
from botocore.model import (
    ListShape,
    MapShape,
    OperationModel,
    ServiceModel,
    Shape,
    StringShape,
    StructureShape,
)
from botocore.paginate import PageIterator

from mypy_boto3_builder.boto3_ports.model import Collection as Boto3Collection
from mypy_boto3_builder.boto3_ports.model import ResourceModel
from mypy_boto3_builder.constants import (
    ATTRIBUTES,
    CLIENT,
    SERVICE_RESOURCE,
)
from mypy_boto3_builder.exceptions import ShapeParserError
from mypy_boto3_builder.logger import get_logger
from mypy_boto3_builder.parsers.resource_loader import ResourceLoader
from mypy_boto3_builder.parsers.shape_parser_types import (
    ActionShape,
    IdentifierShape,
    PaginatorShape,
    PaginatorsShape,
    ResourceShape,
    ResourcesShape,
    SubResourceShape,
    WaitersShape,
)
from mypy_boto3_builder.parsers.typed_dict_map import TypedDictMap
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.structures.argument import Argument
from mypy_boto3_builder.structures.attribute import Attribute
from mypy_boto3_builder.structures.class_record import ClassRecord
from mypy_boto3_builder.structures.method import Method
from mypy_boto3_builder.type_annotations.external_import import ExternalImport
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.internal_import import InternalImport
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.type_constant import TypeConstant
from mypy_boto3_builder.type_annotations.type_literal import TypeLiteral
from mypy_boto3_builder.type_annotations.type_subscript import TypeSubscript
from mypy_boto3_builder.type_annotations.type_typed_dict import TypeTypedDict
from mypy_boto3_builder.type_annotations.type_union import TypeUnion
from mypy_boto3_builder.type_maps.argument_alias_map import get_argument_alias
from mypy_boto3_builder.type_maps.literal_type_map import get_type_literal_stub
from mypy_boto3_builder.type_maps.method_type_map import (
    get_default_value_stub,
    get_method_type_stub,
)
from mypy_boto3_builder.type_maps.required_attribute_map import is_required
from mypy_boto3_builder.type_maps.shape_type_map import (
    get_output_shape_type_stub,
    get_shape_type_stub,
)
from mypy_boto3_builder.type_maps.typed_dicts import (
    EmptyResponseMetadataTypeDef,
    PaginatorConfigTypeDef,
    ResponseMetadataTypeDef,
    WaiterConfigTypeDef,
)
from mypy_boto3_builder.utils.strings import (
    capitalize,
    extract_docstring_from_html,
    get_short_docstring,
    get_type_def_name,
    xform_name,
)
from mypy_boto3_builder.utils.type_checks import is_type_parent, is_union

if TYPE_CHECKING:
    from mypy_boto3_builder.type_annotations.type_parent import TypeParent


class ShapeParser:
    """
    Parser for botocore shape files.

    Arguments:
        service_name -- ServiceName.
    """

    def __init__(self, service_name: ServiceName) -> None:
        self._resource_loader = ResourceLoader()
        service_data = self._resource_loader.get_service_data(service_name)
        self.service_name = service_name
        self.service_model = ServiceModel(service_data, service_name.boto3_name)
        self._resource_name: str = ""
        self._type_literal_map: dict[str, TypeLiteral] = {}
        self._typed_dict_map = TypedDictMap()
        self._output_typed_dict_map = TypedDictMap()
        self._response_typed_dict_map = TypedDictMap()
        self._request_typed_dict_map = TypedDictMap()
        self._fixed_typed_dict_map: dict[TypeTypedDict, TypeTypedDict] = {}

        self.logger = get_logger()

    @property
    def _waiters_shape(self) -> WaitersShape | None:
        return self._resource_loader.load_waiters(self.service_name)

    @property
    def _paginators_shape(self) -> PaginatorsShape | None:
        return self._resource_loader.load_paginators(self.service_name)

    @property
    def _resources_shape(self) -> ResourcesShape | None:
        return self._resource_loader.load_resources(self.service_name)

    @property
    def resource_name(self) -> str:
        """
        Parsed resource name.
        """
        return self._resource_name

    def _get_operation(self, name: str) -> OperationModel:
        return self.service_model.operation_model(name)

    def _get_paginator(self, name: str) -> PaginatorShape:
        if not self._paginators_shape:
            raise ShapeParserError(f"Unknown paginator: {name}")
        try:
            return self._paginators_shape["pagination"][name]
        except KeyError as e:
            raise ShapeParserError(f"Unknown paginator: {name}") from e

    def has_service_resource(self) -> bool:
        """
        Check if service has ServiceResource.
        """
        return self._resource_loader.has_service_resource(self.service_name)

    def _get_service_resource(self) -> ResourceShape:
        if not self._resources_shape:
            raise ShapeParserError("Resource shape not found")
        if "service" not in self._resources_shape:
            raise ShapeParserError("Service resource shape not found")
        return self._resources_shape["service"]

    def get_subresource_names(self) -> list[str]:
        """
        Get available subresource names.
        """
        return list(self._get_subresources().keys())

    def _get_subresources(self) -> dict[str, ResourceShape]:
        if not self._resources_shape:
            return {}
        if "resources" not in self._resources_shape:
            return {}

        return self._resources_shape["resources"]

    def get_service_resource_model(self) -> ResourceModel:
        """
        Get service resource model.
        """
        if not self.has_service_resource():
            raise ShapeParserError("Service resource does not exist")
        return ResourceModel(
            name=self.service_name.boto3_name,
            definition=dict(self._get_service_resource()),
            resource_defs=self._get_subresources(),
        )

    def get_resource_model(self, resource_name: str) -> ResourceModel:
        """
        Get service resource model.
        """
        resource_defs = self._get_subresources()
        if resource_name not in resource_defs:
            raise ShapeParserError(f"Resource {resource_name} not found in subresources")
        return ResourceModel(
            name=resource_name,
            definition=dict(resource_defs[resource_name]),
            resource_defs=resource_defs,
        )

    def _get_resource_shape(self, name: str) -> ResourceShape:
        if name == SERVICE_RESOURCE:
            return self._get_service_resource()

        if not self._resources_shape or "resources" not in self._resources_shape:
            raise ShapeParserError("Resource shape not found")

        try:
            return self._resources_shape["resources"][name]
        except KeyError as e:
            raise ShapeParserError(f"Unknown resource: {name}") from e

    def get_paginator_names(self) -> list[str]:
        """
        Get available paginator names.

        Returns:
            A list of paginator names.
        """
        if not self._paginators_shape or not self._paginators_shape.get("pagination"):
            return []
        result = list(self._paginators_shape["pagination"])
        result.sort()
        return result

    def _parse_arguments(
        self,
        *,
        class_name: str,
        method_name: str,
        operation_name: str,
        shape: StructureShape,
        exclude_names: Iterable[str] = (),
        optional_only: bool = False,
    ) -> list[Argument]:
        result: list[Argument] = []
        required = shape.required_members
        for argument_name, argument_shape in shape.members.items():
            if argument_name in exclude_names:
                continue
            argument_alias = get_argument_alias(self.service_name, operation_name, argument_name)

            argument_type_stub = get_method_type_stub(
                self.service_name,
                class_name,
                method_name,
                argument_name,
            )
            if argument_type_stub is Type.RemoveArgument:
                continue
            if argument_type_stub is not None:
                argument_type = argument_type_stub
            else:
                argument_type = self.parse_shape(argument_shape)
            argument = Argument(argument_alias, argument_type)
            if argument_name not in required:
                argument.default = Type.Ellipsis
            default_value_stub = get_default_value_stub(
                self.service_name,
                class_name,
                method_name,
                argument_name,
            )
            if default_value_stub is not None:
                argument.default = default_value_stub
            if optional_only and argument.required:
                continue

            # FIXME: https://github.com/boto/boto3/issues/2813
            # if not argument.required and argument.type_annotation:
            #     argument.type_annotation = get_optional(argument.type_annotation)

            result.append(argument)

        result.sort(key=lambda x: not x.required)
        return result

    def _parse_return_type(
        self,
        class_name: str,
        method_name: str,
        shape: Shape | None,
    ) -> FakeAnnotation:
        argument_type_stub = get_method_type_stub(
            self.service_name,
            class_name,
            method_name,
            "return",
        )
        if argument_type_stub is not None:
            return argument_type_stub

        if shape:
            return self.parse_shape(shape, is_output=True)

        return Type.none

    @staticmethod
    def _get_kw_flags(method_name: str, arguments: Sequence[Argument]) -> tuple[Argument, ...]:
        if not arguments:
            return ()
        if method_name[:1].isupper():
            return ()

        return (Argument.kwflag(),)

    def get_client_method_map(self) -> dict[str, Method]:
        """
        Get client methods from shape.

        Returns:
            A map of method name to Method.
        """
        self._resource_name = CLIENT
        result: dict[str, Method] = {
            "can_paginate": Method(
                name="can_paginate",
                arguments=(Argument.self(), Argument("operation_name", Type.str)),
                return_type=Type.bool,
            ),
            "generate_presigned_url": Method(
                name="generate_presigned_url",
                arguments=(
                    Argument.self(),
                    Argument("ClientMethod", Type.str),
                    Argument("Params", Type.MappingStrAny, Type.Ellipsis),
                    Argument("ExpiresIn", Type.int, TypeConstant(3600)),
                    Argument("HttpMethod", Type.str, Type.Ellipsis),
                ),
                return_type=Type.str,
            ),
        }
        for operation_name in self.service_model.operation_names:
            operation_model = self._get_operation(operation_name)
            arguments: list[Argument] = [Argument.self()]
            method_name = xform_name(operation_name)

            if operation_model.input_shape is not None:
                shape_arguments = self._parse_arguments(
                    class_name=self._resource_name,
                    method_name=method_name,
                    operation_name=operation_name,
                    shape=operation_model.input_shape,
                )
                arguments.extend(self._get_kw_flags(method_name, shape_arguments))
                arguments.extend(shape_arguments)

            return_type = self._parse_return_type(
                self._resource_name,
                method_name,
                operation_model.output_shape,
            )
            if return_type is Type.none:
                return_type = EmptyResponseMetadataTypeDef

            method = Method(
                name=method_name,
                arguments=arguments,
                return_type=return_type,
                docstring=get_short_docstring(
                    extract_docstring_from_html(operation_model.documentation)
                ),
            )
            if operation_model.input_shape:
                self.create_request_type_annotation(
                    method=method,
                    name=operation_model.input_shape.name,
                    optional_postfix="Request",
                )
            result[method.name] = method
        return result

    def create_request_type_annotation(
        self, method: Method, name: str, optional_postfix: str = "Extra"
    ) -> None:
        """
        Create a request type annotation for the given method and input shape.
        """
        method.create_request_type_annotation(
            self._get_non_clashing_typed_dict_name(
                name,
                optional_postfix=optional_postfix,
            )
        )
        if method.has_request_type_annotation():
            self._request_typed_dict_map.add(method.request_type_annotation)

    def _get_non_clashing_typed_dict_name(self, name: str, optional_postfix: str) -> str:
        temp_typed_dict = TypeTypedDict(get_type_def_name(name), ())
        return self._get_non_clashing_typed_dict_name_for_existing(
            temp_typed_dict,
            optional_postfix=optional_postfix,
        )

    @staticmethod
    def _get_typed_dict_name_for_shape(shape: Shape, postfix: str = "") -> str:
        return get_type_def_name(shape.name, postfix)

    @staticmethod
    def _get_literal_name(shape: StringShape) -> str:
        # FIXME: hack for apigatewayv2
        if shape.name == "__string":
            children_name = "".join(sorted(capitalize(i) for i in shape.enum))
            return f"{children_name}Type"

        name = capitalize(shape.name.lstrip("_"))
        return f"{name}Type"

    def _parse_shape_string(self, shape: StringShape) -> FakeAnnotation:
        if shape.enum:
            literal_name = self._get_literal_name(shape)
            type_literal_stub = get_type_literal_stub(self.service_name, literal_name)
            type_literal = type_literal_stub or TypeLiteral(literal_name, set(shape.enum))
            if literal_name in self._type_literal_map:
                old_type_literal = self._type_literal_map[literal_name]
                if not type_literal.is_same(old_type_literal):
                    raise ShapeParserError(
                        f"Literal {literal_name} has different values:"
                        f" {old_type_literal.children} vs {type_literal.children}",
                    )
                return old_type_literal

            self._type_literal_map[literal_name] = type_literal
            return type_literal

        # FIXME: botocore does not always try to parse response as JSON
        # pattern = shape.metadata.get("pattern", "")
        # if pattern in (
        #     "[\\u0009\\u000A\\u000D\\u0020-\\u00FF]+",
        #     "^[\\u0009\\u000A\\u000D\\u0020-\\u00FF]+$",
        # ):
        #     if output_child:
        #         return Type.DictStrAny
        #     else:
        #         return DictOrStrTypeDef

        return Type.str

    def _parse_shape_map(
        self,
        shape: MapShape,
        *,
        is_output_child: bool = False,
        is_streaming: bool = False,
    ) -> FakeAnnotation:
        parent = Type.Dict if is_output_child else Type.Mapping
        key_child = (
            self.parse_shape(
                shape.key,
                is_output_child=is_output_child,
                is_streaming=is_streaming,
            )
            if shape.key
            else Type.str
        )
        value_child = (
            self.parse_shape(
                shape.value,
                is_output_child=is_output_child,
                is_streaming=is_streaming,
            )
            if shape.value
            else Type.Any
        )
        return TypeSubscript(parent, [key_child, value_child])

    def _get_typed_dict_map(self, *, output: bool, output_child: bool) -> TypedDictMap:
        if output:
            return self._response_typed_dict_map
        if output_child:
            return self._output_typed_dict_map
        return self._typed_dict_map

    def _parse_shape_structure(
        self,
        shape: StructureShape,
        *,
        output: bool = False,
        output_child: bool = False,
        is_streaming: bool = False,
    ) -> FakeAnnotation:
        if not shape.members.items():
            return Type.DictStrAny if output_child else Type.MappingStrAny

        required = shape.required_members
        is_output_or_child = output or output_child
        typed_dict_name = self._get_shape_type_name(shape)
        typed_dict = TypeTypedDict(typed_dict_name)

        typed_dict_map = self._get_typed_dict_map(output=output, output_child=output_child)
        resource_typed_dict_name = self._get_typed_dict_name_for_shape(
            shape, postfix=self.resource_name
        )
        found_typed_dict = typed_dict_map.get(typed_dict.name)
        found_resource_typed_dict = typed_dict_map.get(resource_typed_dict_name)

        if found_resource_typed_dict:
            return found_resource_typed_dict

        typed_dict_map[resource_typed_dict_name] = typed_dict

        for attr_name, attr_shape in shape.members.items():
            typed_dict.add_attribute(
                attr_name,
                self.parse_shape(
                    attr_shape,
                    is_output=False,
                    is_output_child=is_output_or_child,
                    is_streaming=is_streaming,
                ),
                required=attr_name in required,
            )
        if output:
            self._mark_typed_dict_as_total(typed_dict)
            self._add_response_metadata(typed_dict)

        if found_typed_dict and not typed_dict.is_same(found_typed_dict):
            self.logger.debug(
                f"Renaming conflicting {typed_dict.name} to {resource_typed_dict_name}",
                tags=(typed_dict.name, resource_typed_dict_name),
            )
            typed_dict.name = resource_typed_dict_name
        typed_dict_map[typed_dict.name] = typed_dict
        return typed_dict

    def _mark_typed_dict_as_total(self, typed_dict: TypeTypedDict) -> None:
        for attribute in typed_dict.children:
            if is_required(self.service_name, typed_dict.name, attribute.name):
                attribute.mark_as_required()
            else:
                attribute_rendered = attribute.get_type_annotation().render()
                self.logger.debug(
                    f"Leaving output {typed_dict.name}.{attribute.name} as {attribute_rendered}",
                    tags=(typed_dict.name, attribute.name, attribute_rendered),
                )

    def _add_response_metadata(self, typed_dict: TypeTypedDict) -> None:
        child_names = {i.name for i in typed_dict.children}
        if "ResponseMetadata" not in child_names:
            typed_dict.add_attribute(
                "ResponseMetadata",
                ResponseMetadataTypeDef,
                required=True,
            )

    def _parse_shape_list(
        self,
        shape: ListShape,
        *,
        is_output_child: bool = False,
    ) -> FakeAnnotation:
        type_subscript = (
            TypeSubscript(Type.List) if is_output_child else TypeSubscript(Type.Sequence)
        )
        if shape.member:
            type_subscript.add_child(
                self.parse_shape(shape.member, is_output_child=is_output_child),
            )
        else:
            type_subscript.add_child(Type.Any)
        return type_subscript

    def _get_shape_type_name(self, shape: Shape) -> str:
        if isinstance(shape, StructureShape):
            return self._get_typed_dict_name_for_shape(shape)

        if isinstance(shape, StringShape):
            return shape.name

        return shape.type_name

    @staticmethod
    def _get_streaming_body(shape: Shape) -> Shape | None:
        """
        Get the streaming member's shape if any; or None otherwise.
        """
        if not isinstance(shape, StructureShape):
            return None
        payload = shape.serialization.get("payload")
        if payload is not None:
            payload_shape = shape.members.get(payload)
            if isinstance(payload_shape, Shape) and payload_shape.type_name == "blob":
                return payload_shape
        return None

    def _parse_shape_by_type(
        self,
        shape: Shape,
        *,
        is_output_or_child: bool,
        output: bool,
        is_streaming: bool,
    ) -> FakeAnnotation:
        match shape:
            case StringShape():
                return self._parse_shape_string(shape)
            case MapShape():
                return self._parse_shape_map(
                    shape,
                    is_output_child=is_output_or_child,
                    is_streaming=is_streaming,
                )
            case StructureShape():
                return self._parse_shape_structure(
                    shape,
                    output=output,
                    output_child=is_output_or_child,
                    is_streaming=is_streaming,
                )
            case ListShape():
                return self._parse_shape_list(shape, is_output_child=is_output_or_child)
            case _:
                pass

        if shape.type_name in self.get_subresource_names():
            return InternalImport(shape.type_name, use_alias=True)

        self.logger.warning(f"Unknown shape: {shape} {shape.type_name}")
        return Type.Any

    def parse_shape(
        self,
        shape: Shape,
        *,
        is_output: bool = False,
        is_output_child: bool = False,
        is_streaming: bool = False,
    ) -> FakeAnnotation:
        """
        Parse any botocore shape to TypeAnnotation.

        Arguments:
            shape -- Botocore shape.
            output -- Whether shape should use strict output types.
            output_child -- Whether shape parent is marked as output.
            is_streaming -- Whether shape should be streaming.

        Returns:
            TypeAnnotation or similar class.
        """
        if shape.serialization.get("eventstream"):
            shape.serialization["eventstream"] = False
            return TypeSubscript(
                ExternalImport.from_class(EventStream),
                [
                    self.parse_shape(
                        shape,
                        is_output=is_output,
                        is_output_child=is_output_child,
                        is_streaming=is_streaming,
                    ),
                ],
            )
        is_output_or_child = is_output or is_output_child
        if not is_streaming:
            is_streaming = "streaming" in shape.serialization and shape.serialization["streaming"]
            if is_output_or_child:
                is_streaming = self._get_streaming_body(shape) is not None

        type_name = self._get_shape_type_name(shape)
        if is_streaming and shape.type_name == "blob":
            type_name = "blob_streaming"

        if is_output_or_child:
            shape_type_stub = get_output_shape_type_stub(
                self.service_name,
                self._resource_name,
                type_name,
            )
        else:
            shape_type_stub = get_shape_type_stub(self.service_name, self._resource_name, type_name)

        if shape_type_stub:
            return shape_type_stub

        result = self._parse_shape_by_type(
            shape,
            is_output_or_child=is_output_or_child,
            output=is_output,
            is_streaming=is_streaming,
        )
        if isinstance(result, TypeTypedDict):
            replacement = Type.DictStrAny if is_output_or_child else Type.MappingStrAny
            mutated_parents = result.replace_self_references(replacement)
            for mutated_parent in sorted(mutated_parents):
                self.logger.debug(
                    f"Replaced self reference for {result.render()} in {mutated_parent.render()}",
                )

        return result

    def get_paginator_subscript(self, paginator_name: str) -> FakeAnnotation:
        """
        Get Paginator return class.
        """
        operation_name = paginator_name
        operation_shape = self._get_operation(operation_name)
        if operation_shape.output_shape is None:
            return Type.Any

        return self._parse_return_type(
            "Paginator",
            "paginate",
            operation_shape.output_shape,
        )

    def get_paginate_method(self, paginator_name: str) -> Method:
        """
        Get Paginator `paginate` method.

        Arguments:
            paginator_name -- Paginator name.

        Returns:
            Method.
        """
        self._resource_name = "Paginator"
        operation_name = paginator_name
        paginator_shape = self._get_paginator(paginator_name)
        operation_model = self._get_operation(operation_name)
        skip_argument_names: list[str] = []
        input_token: list[str] | str = paginator_shape["input_token"]
        if isinstance(input_token, list):
            skip_argument_names.extend(input_token)
        else:
            skip_argument_names.append(input_token)
        if "limit_key" in paginator_shape:
            skip_argument_names.append(paginator_shape["limit_key"])

        arguments: list[Argument] = [Argument.self()]

        if operation_model.input_shape is not None:
            shape_arguments = self._parse_arguments(
                class_name="Paginator",
                method_name="paginate",
                operation_name=operation_name,
                shape=operation_model.input_shape,
                exclude_names=skip_argument_names,
            )
            shape_arguments.append(
                Argument("PaginationConfig", PaginatorConfigTypeDef, Type.Ellipsis),
            )
            arguments.extend(self._get_kw_flags("paginate", shape_arguments))
            arguments.extend(shape_arguments)

        subscript = self.get_paginator_subscript(operation_name)
        return_type = TypeSubscript(ExternalImport.from_class(PageIterator), [subscript])
        method = Method(
            name="paginate",
            arguments=arguments,
            return_type=return_type,
            type_ignore="override",
        )
        if operation_model.input_shape is not None:
            self.create_request_type_annotation(
                method=method,
                name=operation_model.input_shape.name,
                optional_postfix="Paginate",
            )
        return method

    def get_waiter_names(self) -> list[str]:
        """
        Get Waiter class names.
        """
        if not self._waiters_shape:
            return []

        return list(self._waiters_shape["waiters"].keys())

    def get_wait_method(self, waiter_name: str) -> Method:
        """
        Get Waiter `wait` method.

        Arguments:
            waiter_name -- Waiter name.

        Returns:
            Method.
        """
        self._resource_name = "Waiter"
        if not self._waiters_shape:
            raise ShapeParserError("Waiter shape is not defined")
        operation_name = self._waiters_shape["waiters"][waiter_name]["operation"]
        operation_model = self._get_operation(operation_name)

        arguments: list[Argument] = [Argument.self()]

        if operation_model.input_shape is not None:
            shape_arguments = self._parse_arguments(
                class_name="Waiter",
                method_name="wait",
                operation_name=operation_name,
                shape=operation_model.input_shape,
            )
            shape_arguments.append(Argument("WaiterConfig", WaiterConfigTypeDef, Type.Ellipsis))
            arguments.extend(self._get_kw_flags("wait", shape_arguments))
            arguments.extend(shape_arguments)

        method = Method(
            name="wait",
            arguments=arguments,
            return_type=Type.none,
            type_ignore="override",
        )
        if operation_model.input_shape is not None:
            self.create_request_type_annotation(
                method=method,
                name=operation_model.input_shape.name,
                optional_postfix="Wait",
            )
        return method

    def _get_identifier_type(
        self,
        resource_name: str,
        method_name: str,
        identifier_name: str,
        identifier: IdentifierShape,
    ) -> FakeAnnotation:
        argument_type_stub = get_method_type_stub(
            self.service_name,
            resource_name,
            method_name,
            identifier_name,
        )
        if argument_type_stub:
            return argument_type_stub
        identifier_type = identifier.get("type")
        if identifier_type:
            argument_type_stub = get_shape_type_stub(
                self.service_name,
                resource_name,
                identifier_type,
            )
            if argument_type_stub:
                return argument_type_stub

        return Type.str

    @staticmethod
    def _get_identifier_xform_name(identifier: IdentifierShape) -> str:
        return xform_name(identifier.get("name") or identifier.get("target") or "unknown")

    def _get_identifier_argument(
        self,
        resource_name: str,
        method_name: str,
        identifier: IdentifierShape,
    ) -> Argument:
        argument_name = self._get_identifier_xform_name(identifier)
        argument_type = self._get_identifier_type(
            resource_name,
            method_name,
            argument_name,
            identifier,
        )
        return Argument(argument_name, argument_type)

    def _get_identifier_attribute(
        self,
        resource_name: str,
        identifier: IdentifierShape,
    ) -> Attribute:
        attribute_name = self._get_identifier_xform_name(identifier)
        attribute_type = self._get_identifier_type(
            resource_name,
            ATTRIBUTES,
            attribute_name,
            identifier,
        )
        return Attribute(name=attribute_name, type_annotation=attribute_type, is_identifier=True)

    def _get_existing_sub_resource_names(
        self, service_resource_shape: ResourceShape, names: Iterable[str]
    ) -> set[str]:
        """
        Refer to `boto3.model._get_related_resources` definition.

        Probably it contains a bug that causes the issue with `ec2` package.
        """
        result: set[str] = set()
        has_map = service_resource_shape.get("has", {})
        has_rename_map: dict[str, str] = {}
        for has_name, has_def in has_map.items():
            old_name = has_def.get("resource", {}).get("type")
            if old_name:
                has_rename_map[old_name] = has_name
        for name in names:
            if name not in has_rename_map:
                result.add(name)
            else:
                result.add(has_rename_map[name])

        return result

    def get_service_resource_method_map(self) -> dict[str, Method]:
        """
        Get methods for ServiceResource.

        Returns:
            A map of method name to Method.
        """
        result: dict[str, Method] = {
            "get_available_subresources": Method(
                name="get_available_subresources",
                arguments=[Argument.self()],
                return_type=TypeSubscript(Type.Sequence, [Type.str]),
                docstring="Returns a list of all the available sub-resources for this resource.",
            ),
        }
        self._resource_name = SERVICE_RESOURCE
        service_resource_shape = self._get_resource_shape(self._resource_name)
        for action_name, action_shape in service_resource_shape.get("actions", {}).items():
            method = self._get_resource_method(action_name, action_shape)
            result[method.name] = method

        sub_resource_names = self.get_subresource_names()
        existing_sub_resource_names = self._get_existing_sub_resource_names(
            service_resource_shape, sub_resource_names
        )
        for sub_resource_name in sub_resource_names:
            # FIXME: hack for ec2: KeyPairInfo is not registered, but present in docs
            if sub_resource_name not in existing_sub_resource_names:
                self.logger.debug(
                    f"Skipping {sub_resource_name} sub resource"
                    " because it is not present in ServiceResource.has",
                    tags=sub_resource_name,
                )
                continue
            sub_resource_shape = self._get_resource_shape(sub_resource_name)
            if not self._is_subresource(sub_resource_shape):
                continue
            arguments = [Argument.self()]
            identifiers = sub_resource_shape.get("identifiers", [])
            for identifier in identifiers:
                argument = self._get_identifier_argument(
                    self._resource_name,
                    sub_resource_name,
                    identifier,
                )
                arguments.append(argument)
            method = Method(
                name=sub_resource_name,
                arguments=arguments,
                return_type=InternalImport(sub_resource_name, use_alias=True),
                docstring=f"Creates a {sub_resource_name} resource.",
            )
            result[method.name] = method

        return result

    def get_identifier_attributes(self, resource_name: str) -> list[Attribute]:
        """
        Get attributes for Resource identifiers.

        Arguments:
            resource_name -- Resource name.

        Returns:
            A map of method name to Method.
        """
        self._resource_name = resource_name
        resource_shape = self._get_resource_shape(resource_name)
        identifiers = resource_shape.get("identifiers", [])
        return [
            self._get_identifier_attribute(resource_name, identifier) for identifier in identifiers
        ]

    def get_resource_load_method(self, method_name: str = "load") -> Method:
        """
        Get Resource `load` or `reload` method.
        """
        return Method(name=method_name, arguments=[Argument.self()], return_type=Type.none)

    @staticmethod
    def _is_subresource(subresource_shape: ResourceShape | SubResourceShape) -> bool:
        for identifier in subresource_shape.get("identifiers", []):
            if identifier.get("source") == "data":
                return False

        return True

    def get_resource_method_map(self, resource_name: str) -> dict[str, Method]:
        """
        Get methods for Resource.

        Arguments:
            resource_name -- Resource name.

        Returns:
            A map of method name to Method.
        """
        self._resource_name = resource_name
        resource_shape = self._get_resource_shape(resource_name)
        result: dict[str, Method] = {
            "get_available_subresources": Method(
                name="get_available_subresources",
                arguments=[Argument.self()],
                return_type=TypeSubscript(Type.Sequence, [Type.str]),
                docstring=(
                    f"Returns a list of all the available sub-resources for this {resource_name}."
                ),
            ),
        }

        if "actions" in resource_shape:
            for action_name, action_shape in resource_shape["actions"].items():
                method = self._get_resource_method(action_name, action_shape)
                result[method.name] = method

        if "waiters" in resource_shape:
            for waiter_name in resource_shape["waiters"]:
                state_name = xform_name(waiter_name)
                method = Method(
                    name=f"wait_until_{state_name}",
                    arguments=[Argument.self()],
                    return_type=Type.none,
                    docstring=f"Waits until {resource_name} is {state_name}.",
                )
                result[method.name] = method

        for method_name, sub_resource in resource_shape.get("has", {}).items():
            if "resource" not in sub_resource:
                continue
            data = sub_resource["resource"]
            if not self._is_subresource(data):
                continue
            arguments = [Argument.self()]
            identifiers = data.get("identifiers", [])
            for identifier in identifiers:
                if identifier.get("source") != "input":
                    continue
                argument = self._get_identifier_argument(
                    resource_name,
                    method_name,
                    identifier,
                )
                arguments.append(argument)

            sub_resource_name = data["type"]
            method = Method(
                name=method_name,
                arguments=arguments,
                return_type=InternalImport(sub_resource_name, use_alias=True),
                docstring=f"Creates a {sub_resource_name} resource.",
            )
            result[method.name] = method

        return result

    @staticmethod
    def _get_arg_from_target(target: str) -> str:
        if "[" in target:
            target = target.split("[")[0]
        if "." in target:
            target = target.split(".")[0]
        return target

    def _get_skip_argument_names(self, action_shape: ActionShape) -> set[str]:
        result: set[str] = set()
        params = action_shape["request"].get("params", [])
        for param in params:
            target = param["target"]
            source = param["source"]
            if source == "identifier":
                result.add(self._get_arg_from_target(target))
                continue
            if source == "string" and "." in target:
                result.add(self._get_arg_from_target(target))
                continue

        return result

    def _enrich_arguments_defaults(
        self,
        arguments: list[Argument],
        action_shape: ActionShape,
    ) -> None:
        request = action_shape["request"]
        arguments_map = {a.name: a for a in arguments}
        if "params" in request:
            for param in request["params"]:
                target = param["target"]
                source = param["source"]
                if source == "string" and target in arguments_map:
                    arguments_map[target].default = TypeConstant(param["value"])

    def _get_resource_method(self, action_name: str, action_shape: ActionShape) -> Method:
        return_type: FakeAnnotation = Type.none
        method_name = xform_name(action_name)
        arguments: list[Argument] = [Argument.self()]
        if "resource" in action_shape:
            return_type = self._parse_return_type(
                self.resource_name,
                method_name,
                Shape("resource", action_shape["resource"]),
            )
            path = action_shape["resource"].get("path", "")
            if path.endswith("[]"):
                return_type = TypeSubscript(Type.List, [return_type])

        operation_model = None
        if "request" in action_shape:
            operation_name = action_shape["request"]["operation"]
            operation_model = self._get_operation(operation_name)
            skip_argument_names = self._get_skip_argument_names(action_shape)
            if operation_model.input_shape is not None:
                shape_arguments = self._parse_arguments(
                    class_name=self.resource_name,
                    method_name=method_name,
                    operation_name=operation_name,
                    shape=operation_model.input_shape,
                    exclude_names=skip_argument_names,
                )
                arguments.extend(self._get_kw_flags(method_name, shape_arguments))
                arguments.extend(shape_arguments)

            self._enrich_arguments_defaults(arguments, action_shape)
            arguments.sort(key=lambda x: not x.required)

            if operation_model.output_shape is not None and return_type is Type.none:
                operation_return_type = self.parse_shape(
                    operation_model.output_shape,
                    is_output=True,
                )
                return_type = operation_return_type

        method = Method(
            name=method_name,
            arguments=arguments,
            return_type=return_type,
            docstring=get_short_docstring(
                extract_docstring_from_html(operation_model.documentation),
            ),
        )
        if operation_model.input_shape is not None:
            self.create_request_type_annotation(
                method=method,
                name=operation_model.input_shape.name,
                optional_postfix=f"{self.resource_name}{action_name}",
            )
        return method

    def get_collection_filter_method(
        self,
        name: str,
        collection: Boto3Collection,
        self_type: FakeAnnotation,
    ) -> Method:
        """
        Get `filter` classmethod for Resource collection.

        Arguments:
            name -- Collection record name.
            collection -- Boto3 Collection.
            class_type -- Collection class type annotation.

        Returns:
            Filter Method record.
        """
        result = Method(
            name="filter",
            arguments=[Argument.self()],
            return_type=self_type,
        )
        if not collection.request:
            return result

        operation_name = collection.request.operation
        operation_model = self._get_operation(operation_name)

        if operation_model.input_shape is not None:
            shape_arguments = self._parse_arguments(
                class_name=name,
                method_name=result.name,
                operation_name=operation_name,
                shape=operation_model.input_shape,
                optional_only=True,
            )
            result.arguments.extend(self._get_kw_flags(result.name, shape_arguments))
            result.arguments.extend(shape_arguments)

        return result

    def get_collection_batch_methods(self, name: str, collection: Boto3Collection) -> list[Method]:
        """
        Get batch operations for Resource collection.

        Arguments:
            name -- Collection record name.
            collection -- Boto3 Collection.
            class_type -- Collection self type annotation.

        Returns:
            List of Method records.
        """
        result: list[Method] = []
        for batch_action in collection.batch_actions:
            method = Method(
                name=batch_action.name,
                arguments=[Argument.self()],
                return_type=Type.none,
            )
            result.append(method)
            if batch_action.request:
                operation_name = batch_action.request.operation
                operation_model = self._get_operation(operation_name)
                if operation_model.input_shape is not None:
                    shape_arguments = self._parse_arguments(
                        class_name=name,
                        method_name=batch_action.name,
                        operation_name=operation_name,
                        shape=operation_model.input_shape,
                        optional_only=True,
                    )
                    method.arguments.extend(self._get_kw_flags(batch_action.name, shape_arguments))
                    method.arguments.extend(shape_arguments)
                if operation_model.output_shape is not None:
                    item_return_type = self.parse_shape(
                        operation_model.output_shape,
                        is_output=True,
                    )
                    return_type = TypeSubscript(Type.List, [item_return_type])
                    method.return_type = return_type

        return result

    @staticmethod
    def _get_typed_dict_name_prefix(name: str) -> str:
        if name.endswith("OutputTypeDef"):
            return name[: -len("OutputTypeDef")]
        if name.endswith("TypeDef"):
            return name[: -len("TypeDef")]

        raise ShapeParserError(f"Unknown typed dict name format: {name}")

    def _get_typed_dict(
        self,
        name: str,
        maps: Sequence[dict[str, TypeTypedDict]],
    ) -> TypeTypedDict | None:
        for typed_dict_map in maps:
            if name in typed_dict_map:
                return typed_dict_map[name]
        return None

    def _get_non_clashing_typed_dict_name_for_existing(
        self,
        typed_dict: TypeTypedDict,
        postfix: str = "",
        optional_postfix: str = "Extra",
    ) -> str:
        new_typed_dict_name = get_type_def_name(
            self._get_typed_dict_name_prefix(typed_dict.name),
            postfix,
        )
        clashing_typed_dict = self._get_typed_dict(
            new_typed_dict_name,
            (
                self._typed_dict_map,
                self._output_typed_dict_map,
                self._response_typed_dict_map,
                self._request_typed_dict_map,
            ),
        )
        if not clashing_typed_dict:
            return new_typed_dict_name

        temp_typed_dict = typed_dict.copy()
        temp_typed_dict.name = new_typed_dict_name
        if clashing_typed_dict.is_same(temp_typed_dict):
            return new_typed_dict_name

        self.logger.debug(
            f"Clashing typed dict name found: {new_typed_dict_name}",
            tags=new_typed_dict_name,
        )
        return self._get_non_clashing_typed_dict_name_for_existing(
            temp_typed_dict,
            postfix=optional_postfix,
        )

    def fix_typed_dict_names(self) -> None:
        """
        Fix typed dict names to avoid duplicates.
        """
        output_typed_dict_names = self._output_typed_dict_map.get_sorted_names()
        for name in output_typed_dict_names:
            typed_dict = self._get_typed_dict(
                name,
                (self._typed_dict_map,),
            )
            if typed_dict is None:
                continue

            for output_typed_dict in self._output_typed_dict_map.iterate_by_name(name):
                if typed_dict.is_same(output_typed_dict):
                    continue

                old_typed_dict_name = output_typed_dict.name
                new_typed_dict_name = self._get_non_clashing_typed_dict_name_for_existing(
                    output_typed_dict,
                    "Output",
                )
                self._fixed_typed_dict_map[typed_dict] = output_typed_dict
                self.logger.debug(
                    "Fixing output TypedDict name clash"
                    f" {old_typed_dict_name} -> {new_typed_dict_name}",
                    tags=(old_typed_dict_name, new_typed_dict_name),
                )

                self._output_typed_dict_map.rename(output_typed_dict, new_typed_dict_name)

                if old_typed_dict_name in self._response_typed_dict_map:
                    del self._response_typed_dict_map[old_typed_dict_name]
                    self._response_typed_dict_map.add(output_typed_dict)

        response_typed_dict_names = self._response_typed_dict_map.get_sorted_names()
        for name in response_typed_dict_names:
            typed_dict = self._get_typed_dict(
                name,
                (
                    self._typed_dict_map,
                    self._output_typed_dict_map,
                ),
            )
            if typed_dict is None:
                continue

            for response_typed_dict in self._response_typed_dict_map.iterate_by_name(name):
                if typed_dict.is_same(response_typed_dict):
                    continue

                old_typed_dict_name = response_typed_dict.name
                new_typed_dict_name = self._get_non_clashing_typed_dict_name_for_existing(
                    response_typed_dict,
                    postfix="Response",
                )
                self.logger.debug(
                    "Fixing response TypedDict name clash"
                    f" {old_typed_dict_name} -> {new_typed_dict_name}",
                    tags=(old_typed_dict_name, new_typed_dict_name),
                )

                self._response_typed_dict_map.rename(response_typed_dict, new_typed_dict_name)

    def convert_input_arguments_to_unions(
        self, methods: Iterable[tuple[ClassRecord, Method]]
    ) -> None:
        """
        Accept both input and output shapes in method arguments.

        mypy does not compare TypedDicts, so we need to accept both input and output shapes.
        https://github.com/youtype/mypy_boto3_builder/issues/209
        """
        method_arguments: tuple[tuple[ClassRecord, Method, Argument, TypeParent], ...] = tuple(
            (parent_class, method, argument, argument.type_annotation)
            for parent_class, method in methods
            for argument in method.arguments
            if argument.type_annotation and is_type_parent(argument.type_annotation)
        )
        for input_typed_dict, output_typed_dict in self._fixed_typed_dict_map.items():
            union_name = self._get_non_clashing_typed_dict_name_for_existing(
                input_typed_dict, postfix="Union"
            )
            union_type_annotation = TypeUnion(
                name=union_name,
                children=(input_typed_dict, output_typed_dict),
            )
            for parent_class, method, argument, type_annotation in method_arguments:
                if type_annotation == input_typed_dict:
                    argument.type_annotation = union_type_annotation
                    self.logger.debug(
                        f"Added output shape to {parent_class.name}.{method.name}"
                        f" {argument.name} argument {input_typed_dict.name} ->"
                        f" {union_type_annotation.name}",
                        tags=(
                            parent_class.name,
                            method.name,
                            argument.name,
                            input_typed_dict.name,
                            union_type_annotation.name,
                        ),
                    )

                if type_annotation in self._fixed_typed_dict_map:
                    continue

                sub_parents = type_annotation.find_type_annotation_parents(input_typed_dict)
                for parent in sorted(sub_parents):
                    if is_union(parent) and parent.name == union_name:
                        continue
                    old_parent_render = parent.render()
                    parent.replace_child(input_typed_dict, union_type_annotation)
                    new_parent_render = parent.render()
                    self.logger.debug(
                        f"Added output shape to {parent_class.name}.{method.name}"
                        f" {argument.name} argument sub-type {old_parent_render} ->"
                        f" {new_parent_render}",
                        tags=(
                            parent_class.name,
                            method.name,
                            argument.name,
                            old_parent_render,
                            new_parent_render,
                        ),
                    )
